from minari.data_collector import DataCollector
from minari.data_collector.callbacks import EpisodeMetadataCallback, StepDataCallback
from minari.dataset.episode_data import EpisodeData
from minari.dataset.minari_dataset import MinariDataset
from minari.dataset.step_data import StepData
from minari.storage.hosting import (
    download_dataset,
    list_remote_datasets,
    upload_dataset,
)
from minari.storage.local import delete_dataset, list_local_datasets, load_dataset
from minari.utils import (
    combine_datasets,
    create_dataset_from_buffers,
    get_normalized_score,
    split_dataset,
)


__all__ = [
    # Minari Dataset
    "MinariDataset",
    "EpisodeData",
    "StepData",
    # Data collection
    "DataCollector",
    "EpisodeMetadataCallback",
    "StepDataCallback",
    # Dataset Functions
    "download_dataset",
    "list_remote_datasets",
    "upload_dataset",
    "delete_dataset",
    "list_local_datasets",
    "load_dataset",
    "combine_datasets",
    "create_dataset_from_buffers",
    "split_dataset",
    "get_normalized_score",
]

__version__ = "0.5.0"
supported_dataset_versions = {
    "0.4.0",
    "0.4.1",
    "0.4.2",
    "0.4.3",
    "0.5.0",
}



from __future__ import annotations

import importlib.metadata
import json
import os
import warnings
from typing import Dict, List

from gymnasium import logger

from minari.dataset.minari_dataset import gen_dataset_id, parse_dataset_id
from minari.dataset.minari_storage import METADATA_FILE_NAME
from minari.storage.datasets_root_dir import get_dataset_path
from minari.storage.local import dataset_id_sort_key, load_dataset
from minari.storage.remotes import get_cloud_storage


# Use importlib due to circular import when: "from minari import __version__"
__version__ = importlib.metadata.version("minari")


def upload_dataset(dataset_id: str, key_path: str):
    """Upload a Minari dataset to the remote Farama server.

    If you would like to upload a dataset please first get in touch with the Farama team at contact@farama.org.

    Args:
        dataset_id (str): name id of the local Minari dataset
        key_path (str): path to the credentials file.
    """
    # Avoid circular import
    from minari.namespace import list_remote_namespaces, upload_namespace

    remote_datasets = list_remote_datasets()
    if dataset_id in remote_datasets.keys():
        warnings.warn(
            f"Upload aborted. {dataset_id} is already in remote.", UserWarning
        )
        return

    cloud_storage = get_cloud_storage(key_path=key_path)

    datasets_to_upload = [dataset_id]
    while len(datasets_to_upload):
        dataset_id = datasets_to_upload.pop()
        namespace, _, _ = parse_dataset_id(dataset_id)

        if namespace is not None and namespace not in list_remote_namespaces():
            upload_namespace(namespace, key_path)

        print(f"Uploading dataset {dataset_id}")
        path = get_dataset_path(dataset_id)
        cloud_storage.upload_directory(path, dataset_id)

        dataset = load_dataset(dataset_id)
        combined_datasets = dataset.spec.combined_datasets
        datasets_to_upload.extend(combined_datasets)


def download_dataset(dataset_id: str, force_download: bool = False):
    """Download dataset from remote Farama server.

    An error will be raised if the dataset version is not compatible with the local installed version of Minari.
    This error can be skipped and the download continued with the `force_download` argument. Also, with `force_download`,
    any local datasets that match the id of the downloading dataset will be overridden.

    Args:
        dataset_id (str): name id of the Minari dataset
        force_download (bool): boolean flag for force downloading the dataset. Default Value = False
    """
    # Avoid circular import
    from minari.namespace import create_namespace, list_local_namespaces

    (
        namespace,
        dataset_name,
        dataset_version,
    ) = parse_dataset_id(dataset_id)

    all_dataset_versions = get_remote_dataset_versions(namespace, dataset_name)
    dataset_versionless = gen_dataset_id(namespace, dataset_name)

    # 1. Check if there are any remote versions of the dataset at all
    if not all_dataset_versions:
        raise ValueError(
            f"Couldn't find any version for dataset {dataset_versionless} in the remote Farama server."
        )

    # 2. Check if there are any remote compatible versions with the local installed Minari version
    compatible_dataset_versions = get_remote_dataset_versions(
        namespace,
        dataset_name,
        latest_version=False,
        compatible_minari_version=True,
    )

    if not compatible_dataset_versions:
        message = f"Couldn't find any compatible version of dataset {dataset_versionless} with the local installed version of Minari, {__version__}."
        if not force_download:
            raise ValueError(message)
        else:
            logger.warn(message)

    # 3. Check that the dataset version exists
    if dataset_version not in all_dataset_versions:
        e_msg = f"The dataset version, {dataset_id}, doesn't exist in the remote Farama server."
        if not compatible_dataset_versions:
            raise ValueError(
                e_msg
                + f", and no other compatible versions with the local installed Minari version {__version__} were found."
            )
        else:
            raise ValueError(
                e_msg
                + f" We suggest you download the latest compatible version of this dataset: {dataset_versionless}-v{max(compatible_dataset_versions)}"
            )

    # 4. Check that the dataset version is compatible with the local installed Minari version
    if dataset_version not in compatible_dataset_versions:
        e_msg = (
            f"The version you are trying to download for dataset, {dataset_id}, is not compatible with "
            f"your local installed version of Minari, {__version__}."
        )
        if not force_download:
            raise ValueError(
                e_msg
                + f" You can download the latest compatible version of this dataset: {dataset_versionless}-v{max(compatible_dataset_versions)}."
            )
        # Only a warning and force download incompatible dataset
        elif compatible_dataset_versions:
            logger.warn(
                e_msg
                + f" {dataset_id} will be FORCE download but you can download the latest compatible version of this dataset: {dataset_versionless}-v{max(all_dataset_versions)}."
            )

    # 5. Warning to recommend downloading the latest compatible version of the dataset
    elif dataset_version < max(compatible_dataset_versions):
        logger.warn(
            f"We recommend you install a higher dataset version available and compatible with your local installed Minari version: {dataset_versionless}-v{max(compatible_dataset_versions)}."
        )

    file_path = get_dataset_path(dataset_id)
    if os.path.exists(file_path):
        if not force_download:
            logger.warn(
                f"Skipping Download. Dataset {dataset_id} found locally at {file_path}, Use force_download=True to download the dataset again.\n"
            )
            return

    if namespace is not None and namespace not in list_local_namespaces():
        create_namespace(namespace)

    print(f"\nDownloading {dataset_id} from Farama servers...")
    datasets_path = get_dataset_path("")
    cloud_storage = get_cloud_storage()
    blobs = cloud_storage.list_blobs(prefix=dataset_id)

    for blob in blobs:
        print(f"\n * Downloading data file '{blob.name}' ...\n")
        blob_dir, file_name = os.path.split(blob.name)
        if (
            file_name == ""
        ):  # If the object blob path is a directory continue searching for files
            continue
        blob_local_dir = os.path.join(datasets_path, blob_dir)
        if not os.path.exists(blob_local_dir):
            os.makedirs(blob_local_dir)
        cloud_storage.download_blob(blob, os.path.join(blob_local_dir, file_name))

    print(f"\nDataset {dataset_id} downloaded to {file_path}")

    # Skip a force download of an incompatible dataset version
    if dataset_version in compatible_dataset_versions:
        combined_datasets = load_dataset(dataset_id).spec.combined_datasets

        # If the dataset is a combination of other datasets download the subdatasets recursively
        if len(combined_datasets) > 0:
            print(
                f"\nDataset {dataset_id} is formed by a combination of the following datasets: "
            )
            for name in combined_datasets:
                print(f" * {name}")
            print("\nDownloading extra datasets ...")
            for dataset in combined_datasets:
                download_dataset(dataset_id=dataset)


def list_remote_datasets(
    latest_version: bool = False,
    compatible_minari_version: bool = False,
) -> Dict[str, Dict[str, str]]:
    """Get the names and metadata of all the Minari datasets in the remote Farama server.

    Args:
        latest_version (bool): if `True` only the latest version of the datasets are returned i.e. from ['D4RL/door/human-v0', 'D4RL/door/human-v1`], only the metadata for v1 is returned. Default to `False`.
        compatible_minari_version (bool): if `True` only the datasets compatible with the current Minari version are returned. Default to `False`.

    Returns:
       Dict[str, Dict[str, str]]: keys the names of the Minari datasets and values the metadata
    """
    from minari import supported_dataset_versions

    cloud_storage = get_cloud_storage()
    blobs = cloud_storage.list_blobs()

    # Generate dict = {'dataset_id': (version, metadata)}
    remote_datasets = {}
    for blob in blobs:
        try:
            if os.path.basename(blob.name) == METADATA_FILE_NAME:
                metadata = json.loads(blob.download_as_bytes(client=None))
                if (
                    compatible_minari_version
                    and metadata["minari_version"] not in supported_dataset_versions
                ):
                    continue
                dataset_id = metadata["dataset_id"]
                namespace, dataset_name, version = parse_dataset_id(dataset_id)
                dataset = gen_dataset_id(namespace, dataset_name)

                if latest_version:
                    if (
                        dataset not in remote_datasets
                        or version > remote_datasets[dataset][0]
                    ):
                        remote_datasets[dataset] = (version, metadata)
                else:
                    remote_datasets[dataset_id] = metadata
        except Exception:
            warnings.warn(f"Misconfigured dataset named {blob.name} on remote")

    if latest_version:
        # Convert to dict = {'dataset_id': metadata}
        remote_datasets = dict(
            map(lambda x: (f"{x[0]}-v{x[1][0]}", x[1][1]), remote_datasets.items())
        )

    return {
        k: remote_datasets[k] for k in sorted(remote_datasets, key=dataset_id_sort_key)
    }


def get_remote_dataset_versions(
    namespace: str | None,
    dataset_name: str,
    latest_version: bool = False,
    compatible_minari_version: bool = False,
) -> List[int]:
    """Finds all registered versions in the remote Farama server of the dataset given.

    Args:
        namespace (str | None): identifier for remote namespace. Defaults to None.
        dataset_name (str): name of the dataset.
        latest_version (bool): if `True` only the latest version of the datasets is returned. Default to `False`.
        compatible_minari_version: only return highest version among the datasets compatible with the local installed version of Minari. Default to `False`
    Returns:
        A list of integer versions of the dataset with the specified requirements.
    """
    versions: list[int] = []

    for dataset_id in list_remote_datasets(
        latest_version, compatible_minari_version
    ).keys():
        (
            remote_namespace,
            remote_dataset_name,
            remote_version,
        ) = parse_dataset_id(dataset_id)

        if remote_namespace == namespace and remote_dataset_name == dataset_name:
            versions.append(remote_version)

    return versions



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Union

from minari.dataset.step_data import StepData


@dataclass(frozen=True)
class EpisodeBuffer:
    """Contains the data of a single episode."""

    id: Optional[int] = None
    seed: Optional[int] = None
    options: Optional[dict] = None
    observations: Union[None, list, dict, tuple] = None
    actions: Union[None, list, dict, tuple] = None
    rewards: list = field(default_factory=list)
    terminations: list = field(default_factory=list)
    truncations: list = field(default_factory=list)
    infos: Optional[dict] = None

    def add_step_data(self, step_data: StepData) -> EpisodeBuffer:
        """Add step data dictionary to episode buffer.

        Args:
            step_data (StepData): dictionary with data for a single step

        Returns:
            EpisodeBuffer: episode buffer with appended data
        """
        try:
            import jax.tree_util as jtu
        except ImportError:
            raise ImportError(
                'jax is not installed. Please install it using `pip install "minari[create]"`'
            )

        def _append(data, buffer):
            if isinstance(buffer, list):
                buffer.append(data)
                return buffer
            else:
                return [buffer, data]

        observations = step_data["observation"]
        if self.observations is not None:
            observations = jtu.tree_map(
                _append, step_data["observation"], self.observations
            )
        actions = step_data["action"]
        if self.actions is not None:
            actions = jtu.tree_map(_append, step_data["action"], self.actions)
        infos = step_data["info"]
        if self.infos is not None:
            infos = jtu.tree_map(_append, step_data["info"], self.infos)
        self.rewards.append(step_data["reward"])
        self.terminations.append(step_data["termination"])
        self.truncations.append(step_data["truncation"])

        return EpisodeBuffer(
            id=self.id,
            seed=self.seed,
            options=self.options,
            observations=observations,
            actions=actions,
            rewards=self.rewards,
            terminations=self.terminations,
            truncations=self.truncations,
            infos=infos,
        )

    def __len__(self) -> int:
        """Buffer length."""
        return len(self.rewards)



from __future__ import annotations

import json
import os
import pathlib
import warnings
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Iterable, Optional, Union

import gymnasium as gym
import numpy as np
from gymnasium.envs.registration import EnvSpec

from minari.data_collector.episode_buffer import EpisodeBuffer
from minari.serialization import deserialize_space, serialize_space


PathLike = Union[str, os.PathLike]
METADATA_FILE_NAME = "metadata.json"


class MinariStorage(ABC):
    """Class that handles disk access to the data."""

    FORMAT: str

    def __init__(
        self,
        data_path: pathlib.Path,
        observation_space: gym.Space,
        action_space: gym.Space,
    ):
        self._data_path: pathlib.Path = data_path
        self._observation_space = observation_space
        self._action_space = action_space

    @classmethod
    def read(cls, data_path: PathLike) -> MinariStorage:
        """Create a MinariStorage to read data from a path.

        Args:
            data_path (str or Path): directory where the data is stored.

        Returns:
            A new MinariStorage object to read the data.

        Raises:
            ValueError: if the specified path doesn't exist or doesn't contain any data.
        """
        data_path = pathlib.Path(data_path)
        if not data_path.is_dir():
            raise ValueError(f"The data path {data_path} doesn't exist")
        metadata_file_path = data_path.joinpath(METADATA_FILE_NAME)
        if not metadata_file_path.exists():
            raise ValueError(f"No data found in data path {data_path}")
        with open(metadata_file_path) as file:
            metadata = json.load(file)

        observation_space = None
        action_space = None
        if "observation_space" in metadata.keys():
            serialized_space = metadata["observation_space"]
            assert isinstance(serialized_space, str)
            observation_space = deserialize_space(serialized_space)
        if "action_space" in metadata.keys():
            serialized_space = metadata["action_space"]
            assert isinstance(serialized_space, str)
            action_space = deserialize_space(serialized_space)

        if action_space is None or observation_space is None:
            env_spec_str = metadata.get("env_spec")
            assert isinstance(env_spec_str, str)
            env_spec_str = (  # for gymnasium 1.0.0 compatibility
                env_spec_str.replace('"order_enforce": true,', "")
                .replace('"apply_api_compatibility": false,', "")
                .replace('"autoreset": false, ', "")
            )
            env_spec = EnvSpec.from_json(env_spec_str)
            env = gym.make(env_spec)
            if observation_space is None:
                observation_space = env.observation_space
            if action_space is None:
                action_space = env.action_space

        from minari.dataset._storages import get_minari_storage  # avoid circular import

        return get_minari_storage(metadata["data_format"])(
            data_path,
            observation_space,
            action_space,
        )

    @classmethod
    def new(
        cls,
        data_path: PathLike,
        observation_space: Optional[gym.Space] = None,
        action_space: Optional[gym.Space] = None,
        env_spec: Optional[EnvSpec] = None,
        data_format: str = "hdf5",
    ) -> MinariStorage:
        """Class method to create a new data storage.

        Args:
            data_path (str or Path): directory where the data will be stored.
            observation_space (gymnasium.Space, optional): Gymnasium observation space of the dataset.
            action_space (gymnasium.Space, optional): Gymnasium action space of the dataset.
            env_spec (EnvSpec, optional): Gymnasium EnvSpec of the environment that generates the dataset.
            data_format (str): Format of the data. Default value is "hdf5".

        Returns:
            A new MinariStorage object to write new data.

        Raises:
            ValueError: if the data format is incorrect or the data path already exists.
              Moreover, you need to specify the env_spec or both the observation_space and action_space.
        """
        if env_spec is None and (observation_space is None or action_space is None):
            raise ValueError(
                "Since env_spec is not specified, you need to specify both action space and observation space"
            )
        from minari.dataset._storages import (  # avoid circular import
            get_minari_storage,
            get_storage_keys,
        )

        if data_format not in get_storage_keys():
            raise ValueError(
                f"No storage implemented for {data_format}. Available formats: {get_storage_keys()}"
            )

        data_path = pathlib.Path(data_path)
        data_path.mkdir(exist_ok=True)
        if data_path.joinpath(METADATA_FILE_NAME).exists():
            raise ValueError(
                f"A dataset is already available at {data_path}. Delete it or specify another path"
            )

        if observation_space is None or action_space is None:
            assert env_spec is not None
            env = gym.make(env_spec)
            if observation_space is None:
                observation_space = env.observation_space
            if action_space is None:
                action_space = env.action_space

        metadata: Dict[str, Any] = {
            "total_episodes": 0,
            "total_steps": 0,
            "data_format": data_format,
            "observation_space": serialize_space(observation_space),
            "action_space": serialize_space(action_space),
        }
        if env_spec is not None:
            try:
                metadata["env_spec"] = env_spec.to_json()
            except TypeError as e:
                warnings.warn(
                    f"env_spec is not serializable as {str(e)}. "
                    "You will not be able to recover the environment from the dataset."
                )
        with open(data_path.joinpath(METADATA_FILE_NAME), "w") as f:
            json.dump(metadata, f)

        obj = get_minari_storage(data_format)._create(
            data_path, observation_space, action_space
        )
        return obj

    @classmethod
    @abstractmethod
    def _create(
        cls,
        data_path: pathlib.Path,
        observation_space: gym.Space,
        action_space: gym.Space,
    ) -> MinariStorage:
        ...

    @property
    def metadata(self) -> Dict[str, Any]:
        """Metadata of the dataset."""
        with open(self.data_path.joinpath(METADATA_FILE_NAME)) as file:
            metadata = json.load(file)

        metadata["observation_space"] = self.observation_space
        metadata["action_space"] = self.action_space
        if "author" in metadata:
            metadata["author"] = set(metadata["author"])
        if "author_email" in metadata:
            metadata["author_email"] = set(metadata["author_email"])
        return metadata

    def update_metadata(self, metadata: Dict):
        """Update the metadata adding/modifying some keys.

        Args:
            metadata (dict): dictionary of keys-values to add to the metadata.
        """
        not_updatable_keys = {
            "dataset_id",
            "observation_space",
            "action_space",
            "env_spec",
            "data_format",
            "minari_version",
        }

        assert isinstance(metadata.get("dataset_id", ""), str)
        assert isinstance(metadata.get("total_episodes", 0), int)
        assert isinstance(metadata.get("total_steps", 0), int)
        assert isinstance(metadata.get("code_permalink", ""), str)
        assert isinstance(metadata.get("algorithm_name", ""), str)
        assert isinstance(metadata.get("author", set()), set)
        assert isinstance(metadata.get("author_email", set()), set)
        assert isinstance(metadata.get("minari_version", ""), str)

        with open(self.data_path.joinpath(METADATA_FILE_NAME)) as file:
            saved_metadata = json.load(file)

        forbidden_keys = not_updatable_keys.intersection(metadata.keys())
        forbidden_keys = forbidden_keys.intersection(saved_metadata.keys())

        if forbidden_keys:
            raise ValueError(
                f"You are not allowed to update values for {', '.join(forbidden_keys)}"
            )

        saved_metadata.update(metadata)
        with open(self.data_path.joinpath(METADATA_FILE_NAME), "w") as file:
            json.dump(saved_metadata, file, default=_json_converter)

    @abstractmethod
    def update_episode_metadata(
        self, metadatas: Iterable[Dict], episode_indices: Optional[Iterable] = None
    ):
        """Update the metadata of episodes.

        Args:
            metadatas (Iterable[Dict]): metadatas, one for each episode.
            episode_indices (Iterable, optional): episode indices to update.
            If not specified, all the episodes are considered.
        """
        ...

    @abstractmethod
    def get_episode_metadata(self, episode_indices: Iterable[int]) -> Iterable[Dict]:
        """Get the metadata of episodes.

        Args:
            episode_indices (Iterable[int]): episodes id to return

        Returns:
            metadatas (Iterable[Dict]): episodes metadata
        """
        ...

    def apply(
        self,
        function: Callable[[dict], Any],
        episode_indices: Optional[Iterable] = None,
    ) -> Iterable[Any]:
        """Apply a function to a slice of the data.

        Args:
            function (Callable): function to apply to episodes
            episode_indices (Optional[Iterable]): episodes id to consider

        Returns:
            outs (Iterable): outputs returned by the function applied to episodes
        """
        if episode_indices is None:
            episode_indices = range(self.total_episodes)

        ep_dicts = self.get_episodes(episode_indices)
        return map(function, ep_dicts)

    @abstractmethod
    def get_episodes(self, episode_indices: Iterable[int]) -> Iterable[dict]:
        """Get a list of episodes.

        Args:
            episode_indices (Iterable[int]): episodes id to return

        Returns:
            episodes (Iterable[dict]): episodes data
        """
        ...

    @abstractmethod
    def update_episodes(self, episodes: Iterable[EpisodeBuffer]):
        """Update episodes in the storage from a list of episode buffer.

        Args:
            episodes (Iterable[EpisodeBuffer]): list of episodes buffer.
            They must contain the keys specified in EpsiodeData dataclass, except for `id` which is optional.
            If `id` is specified and exists, the new data is appended to the one in the storage.
        """
        ...

    def update_from_storage(self, storage: MinariStorage):
        """Update the dataset using another MinariStorage.

        Args:
            storage (MinariStorage): the other MinariStorage from which the data will be taken
        """
        for episode in storage.get_episodes(range(storage.total_episodes)):
            episode_buffer = EpisodeBuffer(
                id=None,
                seed=episode.get("seed"),
                observations=episode["observations"],
                actions=episode["actions"],
                rewards=episode["rewards"],
                terminations=episode["terminations"],
                truncations=episode["truncations"],
                infos=episode.get("infos"),
            )
            self.update_episodes([episode_buffer])

        author1 = self.metadata.get("author", set())
        author2 = storage.metadata.get("author", set())
        email1 = self.metadata.get("author_email", set())
        email2 = storage.metadata.get("author_email", set())

        self.update_metadata(
            {
                "author": author1.union(author2),
                "author_email": email1.union(email2),
            }
        )

    def get_size(self) -> float:
        """Returns the dataset size in MB.

        Returns:
            datasize (float): size of the dataset in MB
        """
        datasize = 0
        if self.data_path.exists():
            for filename in self.data_path.glob("**/*"):
                st_size = self.data_path.joinpath(filename).stat().st_size
                datasize += st_size / 1000000

        return np.round(datasize, 1)

    @property
    def data_path(self) -> pathlib.Path:
        """Full path to the dataset."""
        return self._data_path

    @property
    def total_episodes(self) -> int:
        """Total episodes in the dataset."""
        return self.metadata["total_episodes"]

    @property
    def total_steps(self) -> int:
        """Total steps in the dataset."""
        return self.metadata["total_steps"]

    @property
    def observation_space(self) -> gym.Space:
        """Observation Space of the dataset."""
        return self._observation_space

    @property
    def action_space(self) -> gym.Space:
        """Action space of the dataset."""
        return self._action_space


def _json_converter(obj: Any):
    if isinstance(obj, set):
        return list(obj)
    raise TypeError(f"Object of type {type(obj)} is not JSON serializable")



from __future__ import annotations

import pathlib
from collections import OrderedDict
from itertools import zip_longest
from typing import Dict, Iterable, List, Optional, Tuple, Union

import gymnasium as gym
import numpy as np

from minari.data_collector import EpisodeBuffer
from minari.dataset.minari_storage import MinariStorage


try:
    import h5py
except ImportError:
    raise ImportError(
        'h5py is not installed. Please install it using `pip install "minari[hdf5]"`'
    )

_MAIN_FILE_NAME = "main_data.hdf5"


class HDF5Storage(MinariStorage):
    FORMAT = "hdf5"

    def __init__(
        self,
        data_path: pathlib.Path,
        observation_space: gym.Space,
        action_space: gym.Space,
    ):
        super().__init__(data_path, observation_space, action_space)
        file_path = self.data_path.joinpath(_MAIN_FILE_NAME)
        if not file_path.exists():
            raise ValueError(f"No data found in data path {self.data_path}")
        self._file_path = file_path

    @classmethod
    def _create(
        cls,
        data_path: pathlib.Path,
        observation_space: gym.Space,
        action_space: gym.Space,
    ) -> MinariStorage:
        data_path.joinpath(_MAIN_FILE_NAME).touch(exist_ok=False)
        obj = cls(data_path, observation_space, action_space)
        return obj

    def update_episode_metadata(
        self, metadatas: Iterable[Dict], episode_indices: Optional[Iterable] = None
    ):
        if episode_indices is None:
            episode_indices = range(self.total_episodes)

        sentinel = object()
        with h5py.File(self._file_path, "a") as file:
            for metadata, episode_id in zip_longest(
                metadatas, episode_indices, fillvalue=sentinel
            ):
                if sentinel in (metadata, episode_id):
                    raise ValueError(
                        "Metadatas and episode_indices have different lengths"
                    )

                assert isinstance(metadata, dict)
                ep_group = file[f"episode_{episode_id}"]
                ep_group.attrs.update(metadata)

    def get_episode_metadata(self, episode_indices: Iterable[int]) -> Iterable[Dict]:
        with h5py.File(self._file_path, "r") as file:
            for ep_idx in episode_indices:
                ep_group = file[f"episode_{ep_idx}"]
                assert isinstance(ep_group, h5py.Group)
                metadata: dict = dict(ep_group.attrs)
                metadata = unflatten_dict(metadata)
                if metadata.get("seed") is not None:
                    metadata["seed"] = int(metadata["seed"])

                yield metadata

    def _decode_space(
        self,
        hdf_ref: Union[h5py.Group, h5py.Dataset, h5py.Datatype],
        space: gym.spaces.Space,
    ) -> Union[Dict, Tuple, List, np.ndarray]:
        assert not isinstance(hdf_ref, h5py.Datatype)

        if isinstance(space, gym.spaces.Tuple):
            assert isinstance(hdf_ref, h5py.Group)
            result = []
            for i in range(len(hdf_ref.keys())):
                result.append(
                    self._decode_space(hdf_ref[f"_index_{i}"], space.spaces[i])
                )
            return tuple(result)
        elif isinstance(space, gym.spaces.Dict):
            assert isinstance(hdf_ref, h5py.Group)
            result = {}
            for key in hdf_ref.keys():
                result[key] = self._decode_space(hdf_ref[key], space.spaces[key])
            return result
        elif isinstance(space, gym.spaces.Text):
            assert isinstance(hdf_ref, h5py.Dataset)
            result = map(lambda string: string.decode("utf-8"), hdf_ref[()])
            return list(result)
        else:
            assert isinstance(hdf_ref, h5py.Dataset)
            return hdf_ref[()]

    def get_episodes(self, episode_indices: Iterable[int]) -> Iterable[dict]:
        with h5py.File(self._file_path, "r") as file:
            for ep_idx in episode_indices:
                ep_group = file[f"episode_{ep_idx}"]
                assert isinstance(ep_group, h5py.Group)
                infos = None
                if "infos" in ep_group:
                    info_group = ep_group["infos"]
                    assert isinstance(info_group, h5py.Group)
                    infos = _decode_info(info_group)

                ep_dict = {
                    "id": ep_idx,
                    "observations": self._decode_space(
                        ep_group["observations"], self.observation_space
                    ),
                    "actions": self._decode_space(
                        ep_group["actions"], self.action_space
                    ),
                    "infos": infos,
                }
                for key in {"rewards", "terminations", "truncations"}:
                    group_value = ep_group[key]
                    assert isinstance(group_value, h5py.Dataset)
                    ep_dict[key] = group_value[:]

                yield ep_dict

    def update_episodes(self, episodes: Iterable[EpisodeBuffer]):
        additional_steps = 0
        with h5py.File(self._file_path, "a", track_order=True) as file:
            for eps_buff in episodes:
                total_episodes = len(file.keys())
                episode_id = eps_buff.id if eps_buff.id is not None else total_episodes
                assert (
                    episode_id <= total_episodes
                ), "Invalid episode id; ids must be sequential."
                episode_group = _get_from_h5py(file, f"episode_{episode_id}")
                episode_group.attrs["id"] = episode_id
                if eps_buff.seed is not None:
                    assert "seed" not in episode_group.attrs.keys()
                    episode_group.attrs["seed"] = eps_buff.seed
                if eps_buff.options is not None:
                    assert "options" not in episode_group.attrs.keys()
                    flatten_option = flatten_dict(eps_buff.options, "options")
                    episode_group.attrs.update(flatten_option)

                episode_steps = len(eps_buff.rewards)
                episode_group.attrs["total_steps"] = episode_steps
                additional_steps += episode_steps

                dict_buffer = {
                    "observations": eps_buff.observations,
                    "actions": eps_buff.actions,
                    "rewards": eps_buff.rewards,
                    "terminations": eps_buff.terminations,
                    "truncations": eps_buff.truncations,
                    "infos": eps_buff.infos,
                }
                _add_episode_to_group(dict_buffer, episode_group)

            total_episodes = len(file.keys())

        total_steps = self.total_steps + additional_steps
        self.update_metadata(
            {"total_steps": total_steps, "total_episodes": total_episodes}
        )


def _get_from_h5py(group: h5py.Group, name: str) -> h5py.Group:
    if name in group:
        subgroup = group.get(name)
        assert isinstance(subgroup, h5py.Group)
    else:
        subgroup = group.create_group(name)

    return subgroup


def _add_episode_to_group(episode_buffer: Dict, episode_group: h5py.Group):
    for key, data in episode_buffer.items():
        if isinstance(data, dict):
            episode_group_to_clear = _get_from_h5py(episode_group, key)
            _add_episode_to_group(data, episode_group_to_clear)
        elif isinstance(data, tuple):
            dict_data = {f"_index_{i}": subdata for i, subdata in enumerate(data)}
            episode_group_to_clear = _get_from_h5py(episode_group, key)
            _add_episode_to_group(dict_data, episode_group_to_clear)
        elif isinstance(data, List) and all(
            isinstance(entry, OrderedDict) for entry in data
        ):  # list of OrderedDict
            dict_data = {key: [entry[key] for entry in data] for key in data[0].keys()}
            episode_group_to_clear = _get_from_h5py(episode_group, key)
            _add_episode_to_group(dict_data, episode_group_to_clear)

        # leaf data
        elif key in episode_group:
            dataset = episode_group[key]
            assert isinstance(dataset, h5py.Dataset)
            dataset.resize((dataset.shape[0] + len(data), *dataset.shape[1:]))
            dataset[-len(data) :] = data
        elif not isinstance(data, Iterable):
            if data is not None:
                episode_group.create_dataset(key, data=data)
        else:
            dtype = None
            if all(map(lambda elem: isinstance(elem, str), data)):
                dtype = h5py.string_dtype(encoding="utf-8")
            dshape = ()
            if hasattr(data[0], "shape"):
                dshape = data[0].shape

            episode_group.create_dataset(
                key, data=data, dtype=dtype, chunks=True, maxshape=(None, *dshape)
            )


def _decode_info(info_group: h5py.Group) -> Dict:
    result = {}
    for key, value in info_group.items():
        if isinstance(value, h5py.Group):
            result[key] = _decode_info(value)
        elif isinstance(value, h5py.Dataset):
            result[key] = value[()]
        else:
            raise ValueError(
                "Infos are in an unsupported format; see Minari documentation for supported formats."
            )
    return result


def flatten_dict(d: Dict, parent_key: str) -> Dict:
    flatten_d = {}
    for k, v in d.items():
        new_key = f"{parent_key}/{k}"
        if isinstance(v, dict):
            flatten_d.update(flatten_dict(v, new_key))
        else:
            flatten_d[new_key] = v
    return flatten_d


def unflatten_dict(d: Dict) -> Dict:
    result = {}
    for k, v in d.items():
        keys = k.split("/")
        current = result
        for key in keys[:-1]:
            if key not in current:
                current[key] = {}
            current = current[key]
        current[keys[-1]] = v
    return result



import importlib.metadata
import os
import shutil
import warnings
from typing import Dict, Iterable, Tuple, Union

from minari.dataset.minari_dataset import (
    MinariDataset,
    gen_dataset_id,
    parse_dataset_id,
)
from minari.dataset.minari_storage import MinariStorage
from minari.storage import hosting
from minari.storage.datasets_root_dir import get_dataset_path


# Use importlib due to circular import when: "from minari import __version__"
__version__ = importlib.metadata.version("minari")


def list_non_hidden_dirs(path: str) -> Iterable[str]:
    """List all non-hidden subdirectories."""
    for d in os.scandir(path):
        if d.is_dir() and (not d.name.startswith(".")):
            yield d.name


def dataset_id_sort_key(dataset_id: str) -> Tuple[str, str, int]:
    """Key for sorting dataset ids first by namespace, and then alphabetically."""
    namespace, dataset_name, version = parse_dataset_id(dataset_id)
    namespace = "" if namespace is None else namespace
    return (namespace, dataset_name, version)


def load_dataset(dataset_id: str, download: bool = False):
    """Retrieve Minari dataset from local database.

    Args:
        dataset_id (str): name id of Minari dataset
        download (bool): if `True` download the dataset if it is not found locally. Default to `False`.

    Returns:
        MinariDataset
    """
    file_path = get_dataset_path(dataset_id)
    data_path = os.path.join(file_path, "data")

    if not os.path.exists(data_path):
        if not download:
            raise FileNotFoundError(
                f"Dataset {dataset_id} not found locally at {file_path}. Use download=True to download the dataset."
            )

        hosting.download_dataset(dataset_id)

    return MinariDataset(data_path)


def list_local_datasets(
    latest_version: bool = False,
    compatible_minari_version: bool = False,
) -> Dict[str, Dict[str, Union[str, int, bool]]]:
    """Get the ids and metadata of all the Minari datasets in the local database.

    Args:
        latest_version (bool): if `True` only the latest version of the datasets are returned i.e. from ['D4RL/door/human-v0', 'D4RL/door/human-v1`], only the metadata for v1 is returned. Default to `False`.
        compatible_minari_version (bool): if `True` only the datasets compatible with the current Minari version are returned. Default to `False`.

    Returns:
       Dict[str, Dict[str, str]]: keys the names of the Minari datasets and values the metadata
    """
    from minari import supported_dataset_versions

    datasets_path = get_dataset_path("")
    dataset_ids = []

    def recurse_directories(base_path, namespace):
        parent_dir = os.path.join(base_path, namespace)
        for dir_name in list_non_hidden_dirs(parent_dir):
            dir_path = os.path.join(parent_dir, dir_name)
            namespaced_dir_name = os.path.join(namespace, dir_name)
            # Minari datasets must contain the data directory.
            if "data" in os.listdir(dir_path):
                dataset_ids.append(namespaced_dir_name)
            else:
                recurse_directories(base_path, namespaced_dir_name)

    recurse_directories(datasets_path, "")

    dataset_ids = sorted(dataset_ids, key=dataset_id_sort_key)

    local_datasets = {}
    for dst_id in dataset_ids:
        data_path = os.path.join(datasets_path, dst_id, "data")
        try:
            metadata = MinariStorage.read(data_path).metadata
            metadata_id = metadata["dataset_id"]

            if dst_id != metadata_id:
                raise ValueError(
                    f"Namespace location '{dst_id}' does not match id '{metadata_id}'."
                )
        except Exception as e:
            warnings.warn(f"Misconfigured dataset named {dst_id}: {e}")
            continue

        if (
            compatible_minari_version
            and metadata["minari_version"] not in supported_dataset_versions
        ):
            continue

        namespace, dataset_name, version = parse_dataset_id(dst_id)
        dataset = gen_dataset_id(namespace, dataset_name)

        if latest_version:
            if dataset not in local_datasets or version > local_datasets[dataset][0]:
                local_datasets[dataset] = (version, metadata)
        else:
            local_datasets[dst_id] = metadata

    if latest_version:
        # Return dict = {'dataset_id': metadata}
        return dict(
            map(lambda x: (f"{x[0]}-v{x[1][0]}", x[1][1]), local_datasets.items())
        )
    else:
        return local_datasets


def delete_dataset(dataset_id: str):
    """Delete a Minari dataset from the local Minari database.

    Args:
        dataset_id (str): name id of the Minari dataset
    """
    dataset_path = get_dataset_path(dataset_id)
    shutil.rmtree(dataset_path)
    print(f"Dataset {dataset_id} deleted!")



from typing import Any, Dict, Optional

import gymnasium as gym

from minari.dataset.step_data import StepData


class StepDataCallback:
    """Callback to create step data dictionary from the return data of each Gymnasium environment step.

    This callback can be overridden to add extra environment information in each step or
    edit the observation, action, reward, termination, truncation, or info returns.
    """

    def __call__(
        self,
        env: gym.Env,
        obs: Any,
        info: Dict[str, Any],
        action: Optional[Any] = None,
        rew: Optional[Any] = None,
        terminated: Optional[bool] = None,
        truncated: Optional[bool] = None,
    ) -> StepData:
        """Callback method.

        The input arguments belong to a Gymnasium stepping transition: `obs, rew, terminated, truncated, info = env.step(action)`.
        Override this method to add additional keys or edit each environment's step returns. Additional nested dictionaries can be added to the returned step dictionary
        as follows:

        .. code::

            class CustomStepDataCallback(StepDataCallback):
                def __call__(self, env, **kwargs):
                    step_data = super().__call__(env, **kwargs)
                    step_data['environment_states'] = {}
                    step_data['environment_states']['pose'] = {}
                    step_data['environment_states']['pose']['position'] = env.position
                    step_data['environment_states']['pose']['orientation'] = env.orientation
                    step_data['environment_states']['velocity'] = env.velocity

                    return step_data

        The Minari dataset will contain a dictionary called `environment_states` with `velocity` value and another dictionary `pose`
        with `position` and `orientation`

        Args:
            env (gym.Env): current Gymnasium environment.
            obs (Any): observation returned by `env.step(action)`
            info (Dict): information dictionary returned by `env.step(action)`
            action (Optional[Any], optional): stepping action in `env.step(action)`. Defaults to None.
            rew (Optional[Any], optional): reward returned by `env.step(action)`. Defaults to None.
            terminated (Optional[Any], optional): terminated returned by `env.step(action)`. Defaults to None.
            truncated (Optional[Any], optional): truncated returned by `env.step(action)`. Defaults to None.

        Returns:
            Dict: dictionary step data. Must contain the keys in STEP_DATA_KEYS = {'actions', 'observations',
                    'rewards', 'terminations', 'truncations', 'infos'}. Additional key's can be added with nested dictionaries
        """
        step_data: StepData = {
            "action": action,
            "observation": obs,
            "reward": rew,
            "termination": terminated,
            "truncation": truncated,
            "info": info,
        }

        return step_data



from typing import Dict


class EpisodeMetadataCallback:
    """Callback to full episode after saving to hdf5 file as a group.

    This callback can be overridden to add extra metadata attributes or statistics to
    each episode in the Minari dataset. The custom callback can then be
    passed to the DataCollector wrapper to the `episode_metadata_callback` argument.
    """

    def __call__(self, episode: Dict) -> Dict:
        """Callback method.

        Override this method to add custom attribute metadata to the episode group.

        Args:
            episode (dict): the dict that contains an episode's data
        Returns:
            Dict: episode metadata.
        """
        return {
            "rewards_sum": float(episode["rewards"].sum()),
            "rewards_mean": float(episode["rewards"].mean()),
            "rewards_std": float(episode["rewards"].std()),
            "rewards_max": float(episode["rewards"].max()),
            "rewards_min": float(episode["rewards"].min()),
        }



from typing import Any, Dict, Optional, SupportsFloat, TypedDict


class StepData(TypedDict):
    """Object containing data of a single environment step."""

    observation: Any
    action: Optional[Any]
    reward: Optional[SupportsFloat]
    termination: Optional[bool]
    truncation: Optional[bool]
    info: Dict[str, Any]



import pytest

from minari.serialization import deserialize_space, serialize_space
from tests.common import test_spaces, unsupported_test_spaces


@pytest.mark.parametrize("space", test_spaces)
def test_space_serialize_deserialize(space):
    space_str = serialize_space(space)
    reconstructed_space = deserialize_space(space_str)
    reserialized_space_str = serialize_space(reconstructed_space)
    assert space_str == reserialized_space_str

    space.seed(0)
    reconstructed_space.seed(0)
    action_1 = space.sample()
    action_2 = reconstructed_space.sample()
    assert space.contains(action_2)
    assert reconstructed_space.contains(action_1)


@pytest.mark.parametrize("space", unsupported_test_spaces)
def test_space_serialize_deserialize_unsupported(space):
    with pytest.raises(
        NotImplementedError, match=r"No serialization method available for .+"
    ):
        serialize_space(space)



from dataclasses import dataclass
from typing import Any

import numpy as np


@dataclass(frozen=True)
class EpisodeData:
    """Contains the datasets data for a single episode."""

    id: int
    observations: Any
    actions: Any
    rewards: np.ndarray
    terminations: np.ndarray
    truncations: np.ndarray
    infos: dict

    def __len__(self) -> int:
        return len(self.rewards)

    def __repr__(self) -> str:
        return (
            "EpisodeData("
            f"id={self.id}, "
            f"total_steps={len(self)}, "
            f"observations={EpisodeData._repr_space_values(self.observations)}, "
            f"actions={EpisodeData._repr_space_values(self.actions)}, "
            f"rewards=ndarray of {len(self.rewards)} floats, "
            f"terminations=ndarray of {len(self.terminations)} bools, "
            f"truncations=ndarray of {len(self.truncations)} bools, "
            f"infos=dict with the following keys: {list(self.infos.keys())}"
            ")"
        )

    @staticmethod
    def _repr_space_values(value):
        if isinstance(value, np.ndarray):
            return f"ndarray of shape {value.shape} and dtype {value.dtype}"
        elif isinstance(value, dict):
            reprs = [
                f"{k}: {EpisodeData._repr_space_values(v)}" for k, v in value.items()
            ]
            dict_repr = ", ".join(reprs)
            return "{" + dict_repr + "}"
        elif isinstance(value, tuple):
            reprs = [EpisodeData._repr_space_values(v) for v in value]
            values_repr = ", ".join(reprs)
            return "(" + values_repr + ")"
        else:
            return repr(value)



from __future__ import annotations

import json
from collections import defaultdict
from functools import singledispatch
from typing import Dict, Union

import numpy as np
from gymnasium import spaces


@singledispatch
def serialize_space(space: spaces.Space, to_string=True) -> Union[Dict, str]:
    raise NotImplementedError(f"No serialization method available for {space}")


@serialize_space.register(spaces.Box)
def _serialize_box(space: spaces.Box, to_string=True) -> Union[Dict, str]:
    result = {}
    result["type"] = "Box"
    result["dtype"] = str(space.dtype)
    result["shape"] = list(space.shape)
    # we have to use python float type to serialize the np.float32 types
    result["low"] = space.low.tolist()
    result["high"] = space.high.tolist()

    if to_string:
        result = json.dumps(result)
    return result


@serialize_space.register(spaces.Discrete)
def _serialize_discrete(space: spaces.Discrete, to_string=True) -> Union[Dict, str]:
    result = {}
    result["type"] = "Discrete"
    result["dtype"] = "int64"  # this seems to be hardcoded in Gymnasium
    # we need to cast from np.int64 to python's int type in order to serialize
    result["start"] = int(space.start)
    result["n"] = int(space.n)

    if to_string:
        result = json.dumps(result)
    return result


@serialize_space.register(spaces.Dict)
def _serialize_dict(space: spaces.Dict, to_string=True) -> Union[Dict, str]:
    result = {"type": "Dict", "subspaces": {}}
    for key in space.spaces.keys():
        result["subspaces"][key] = serialize_space(space.spaces[key], to_string=False)

    if to_string:
        result = json.dumps(result)
    return result


@serialize_space.register(spaces.Tuple)
def _serialize_tuple(space: spaces.Tuple, to_string=True) -> Union[Dict, str]:
    result = {"type": "Tuple", "subspaces": []}
    for subspace in space.spaces:
        result["subspaces"].append(serialize_space(subspace, to_string=False))

    if to_string:
        return json.dumps(result)
    else:
        return result


@serialize_space.register(spaces.Text)
def _serialize_text(space: spaces.Text, to_string=True) -> Union[Dict, str]:
    result = {
        "type": "Text",
        "max_length": space.max_length,
        "min_length": space.min_length,
        "charset": space.characters,
    }

    if to_string:
        return json.dumps(result)
    else:
        return result


class type_value_dispatch:
    def __init__(self, func) -> None:
        self.registry = defaultdict(func)

    def register(self, type: str):
        def decorator(method):
            self.registry[type] = method
            return method

        return decorator

    def __call__(self, space_dict: Union[Dict, str]) -> spaces.Space:
        if not isinstance(space_dict, Dict):
            space_dict = json.loads(space_dict)

        assert isinstance(space_dict, Dict)
        return self.registry[space_dict["type"]](space_dict)


@type_value_dispatch
def deserialize_space(space_dict: Dict) -> spaces.Space:
    raise NotImplementedError(
        f"No deserialization method available for {space_dict['type']}"
    )


@deserialize_space.register("Tuple")
def _deserialize_tuple(space_dict: Dict) -> spaces.Tuple:
    assert space_dict["type"] == "Tuple"
    subspaces = tuple(
        deserialize_space(subspace) for subspace in space_dict["subspaces"]
    )
    return spaces.Tuple(subspaces)


@deserialize_space.register("Dict")
def _deserialize_dict(space_dict: Dict) -> spaces.Dict:
    assert space_dict["type"] == "Dict"
    subspaces = {
        key: deserialize_space(space_dict["subspaces"][key])
        for key in space_dict["subspaces"]
    }
    return spaces.Dict(subspaces)


@deserialize_space.register("Box")
def _deserialize_box(space_dict: Dict) -> spaces.Box:
    assert space_dict["type"] == "Box"
    shape = tuple(space_dict["shape"])
    dtype = np.dtype(space_dict["dtype"])
    low = np.array(space_dict["low"], dtype=dtype)
    high = np.array(space_dict["high"], dtype=dtype)
    return spaces.Box(low=low, high=high, shape=shape, dtype=dtype)  # type: ignore


@deserialize_space.register("Discrete")
def _deserialize_discrete(space_dict: Dict) -> spaces.Discrete:
    assert space_dict["type"] == "Discrete"
    n = space_dict["n"]
    start = space_dict["start"]
    return spaces.Discrete(n=n, start=start)


@deserialize_space.register("Text")
def _deserialize_text(space_dict: Dict) -> spaces.Text:
    assert space_dict["type"] == "Text"
    return spaces.Text(
        max_length=space_dict["max_length"],
        min_length=space_dict["min_length"],
        charset=space_dict["charset"],
    )



import copy
import json
import os
import re
import warnings
from typing import Any, Dict, List, Optional

from minari.storage import get_dataset_path
from minari.storage.hosting import get_cloud_storage
from minari.storage.local import list_non_hidden_dirs


NAMESPACE_REGEX = re.compile(r"[-_\w][-_\w/]*[-_\w]+")
NAMESPACE_METADATA_FILENAME = "namespace_metadata.json"


def create_namespace(
    namespace: str,
    description: Optional[str] = None,
    **kwargs,
) -> None:
    """Create a local namespace.

    Namespaces are a directory-like structure that can contain multiple datasets, or
    other namespaces. Namespaces are prepended onto a ``dataset_id`` with a forward
    slash. For example a dataset with id ``cartpole/test-v0`` resides in
    the ``cartpole`` namespace. Namespaces can be nested.

    Note: The namespace API is an experimental feature and may change in future releases.

    Args:
        namespace (str): identifier for namespace created/updated.
        description (str | None): string metadata describing the namespace. Defaults to None.
        **kwargs: any other metadata in addition to the description.
    """
    validate_namespace(namespace)

    if namespace in list_local_namespaces():
        raise ValueError(f"Namespace '{namespace}' already exists.")

    metadata = copy.deepcopy(kwargs)
    if description is not None:
        metadata["description"] = description
    directory = get_dataset_path(namespace)
    os.makedirs(directory, exist_ok=True)

    with open(directory / NAMESPACE_METADATA_FILENAME, "w") as file:
        json.dump(metadata, file)


def update_namespace_metadata(
    namespace: str,
    description: Optional[str] = None,
    **kwargs,
) -> None:
    """Update an existing local namespace, overwriting existing namespace metadata.

    Note: The namespace API is an experimental feature and may change in future releases.

    Args:
        namespace (str): identifier for namespace created/updated.
        description (str | None): string metadata describing the namespace. Defaults to None.
        **kwargs: any other metadata in addition to the description.
    """
    validate_namespace(namespace)

    if namespace not in list_local_namespaces():
        raise ValueError(f"Namespace {namespace} does not exist locally.")

    metadata = copy.deepcopy(kwargs)
    metadata["description"] = description
    metadata_filepath = get_dataset_path(namespace) / NAMESPACE_METADATA_FILENAME

    with open(metadata_filepath, "w") as file:
        json.dump(metadata, file)


def get_namespace_metadata(namespace: str) -> Dict[str, Any]:
    """Load local namespace metadata.

    Note: The namespace API is an experimental feature and may change in future releases.

    Args:
        namespace (str): identifier for local namespace.

    Returns:
        Dict[str, Any]: metadata dict.
    """
    validate_namespace(namespace)

    if namespace not in list_local_namespaces():
        raise ValueError(f"Namespace {namespace} does not exist locally.")

    metadata_filepath = get_dataset_path(namespace) / NAMESPACE_METADATA_FILENAME

    with open(metadata_filepath) as file:
        metadata = json.load(file)

    return metadata


def delete_namespace(namespace: str) -> None:
    """Delete local namespace. Only empty namespaces can be deleted.

    Note: The namespace API is an experimental feature and may change in future releases.

    Args:
        namespace (str): identifier for local namespace.
    """
    validate_namespace(namespace)

    if namespace not in list_local_namespaces():
        warnings.warn(f"Namespace '{namespace}' does not exist.", UserWarning)
        return

    directory = get_dataset_path(namespace)
    assert os.path.isdir(directory)
    dir_contents = os.listdir(directory)
    has_metadata = NAMESPACE_METADATA_FILENAME in dir_contents

    if len(dir_contents) != int(has_metadata):
        raise ValueError(
            f"Namespace {directory} is not empty. All datasets must be deleted first."
        )

    if has_metadata:
        os.remove(directory / NAMESPACE_METADATA_FILENAME)

    os.rmdir(directory)


def list_local_namespaces() -> List[str]:
    """Get the names of the namespaces in the local database.

    Note: The namespace API is an experimental feature and may change in future releases.

    Returns:
       List[str]: names of all local namespaces.
    """
    datasets_path = get_dataset_path("")
    namespaces = []

    def recurse_directories(base_path, namespace):
        parent_dir = os.path.join(base_path, namespace)
        for dir_name in list_non_hidden_dirs(parent_dir):
            dir_path = os.path.join(parent_dir, dir_name)
            namespaced_dir_name = os.path.join(namespace, dir_name)
            dir_contents = os.listdir(dir_path)

            if NAMESPACE_METADATA_FILENAME in dir_contents:
                namespaces.append(namespaced_dir_name)

            # Don't recurse the subdirectories of a Minari dataset
            if "data" not in dir_contents:
                recurse_directories(base_path, namespaced_dir_name)

    recurse_directories(datasets_path, "")

    return sorted(namespaces)


def list_remote_namespaces() -> List[str]:
    """Get the names of the namespaces in the remote server.

    Note: The namespace API is an experimental feature and may change in future releases.

    Returns:
       List[str]: names of all remote namespaces.
    """
    cloud_storage = get_cloud_storage()
    blobs = cloud_storage.list_blobs()

    remote_namespaces = []

    for blob in blobs:
        if os.path.basename(blob.name) == NAMESPACE_METADATA_FILENAME:
            namespace = os.path.dirname(blob.name)
            remote_namespaces.append(namespace)

    return remote_namespaces


def download_namespace_metadata(namespace: str, overwrite: bool = False) -> None:
    """Download remote namespace to local database.

    Note: The namespace API is an experimental feature and may change in future releases.

    Args:
        namespace (str): identifier for the remote namespace.
        overwrite (bool): whether to overwrite existing local metadata. Defaults to False.
    """
    validate_namespace(namespace)

    if namespace not in list_remote_namespaces():
        raise ValueError(
            f"The namespace '{namespace}' doesn't exist in the remote Farama server."
        )

    metadata_filename = os.path.join(namespace, NAMESPACE_METADATA_FILENAME)
    cloud_storage = get_cloud_storage()
    blobs = list(cloud_storage.list_blobs(prefix=metadata_filename))

    if not blobs:
        return

    assert len(blobs) == 1

    if overwrite or namespace not in list_local_namespaces():
        directory = get_dataset_path(namespace)
        os.makedirs(directory, exist_ok=True)
        blobs[0].download_to_filename(directory / NAMESPACE_METADATA_FILENAME)
    else:
        warnings.warn(
            f"Skipping update for namespace '{namespace}' due to existing local metadata. Set overwrite=True to overwrite local data.",
            UserWarning,
        )


def upload_namespace(namespace: str, key_path: str) -> None:
    """Upload a local namespace to the remote server.

    If you would like to upload a namespace please first get in touch with the Farama team at contact@farama.org.

    Note: The namespace API is an experimental feature and may change in future releases.

    Args:
        namespace (str): identifier for the local namespace.
        key_path (str): path to the credentials file.
    """
    validate_namespace(namespace)

    if namespace not in list_local_namespaces():
        raise ValueError(f"Namespace '{namespace}' does not exist locally.")

    if namespace in list_remote_namespaces():
        warnings.warn(
            f"Upload aborted. Namespace '{namespace}' is already in remote.",
            UserWarning,
        )
        return

    cloud_storage = get_cloud_storage(key_path=key_path)

    print(f"Uploading namespace '{namespace}'")

    local_filepath = get_dataset_path(namespace) / NAMESPACE_METADATA_FILENAME
    remote_filepath = os.path.join(namespace, NAMESPACE_METADATA_FILENAME)
    cloud_storage.upload_file(local_filepath, remote_filepath)


def validate_namespace(namespace: Optional[str]) -> None:
    """Validate a namespace identifier.

    Note: The namespace API is an experimental feature and may change in future releases.

    Args:
        namespace (str | None): identifier to validate
    """
    if namespace is None:
        raise TypeError("Namespace cannot be None")

    if not NAMESPACE_REGEX.fullmatch(namespace):
        raise ValueError(f"Malformed namespace: {namespace}")



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional


class CloudStorage(ABC):
    @abstractmethod
    def __init__(self, name: str, key_path: Optional[str]) -> None:
        ...

    @abstractmethod
    def upload_directory(self, path: Path, remote_dir_path: str) -> None:
        ...

    @abstractmethod
    def upload_file(self, local_path: Path, remote_path: str) -> None:
        ...

    @abstractmethod
    def list_blobs(self, prefix: Optional[str] = None) -> list:
        ...

    @abstractmethod
    def download_blob(self, blob: Any, file_path: str) -> None:
        ...



from __future__ import annotations

import copy
import logging
import os
import secrets
import shutil
import tempfile
from typing import Any, Callable, Dict, Optional, SupportsFloat, Type

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType
from gymnasium.envs.registration import EnvSpec

from minari.data_collector.callbacks import EpisodeMetadataCallback, StepDataCallback
from minari.data_collector.episode_buffer import EpisodeBuffer
from minari.dataset.minari_dataset import MinariDataset, parse_dataset_id
from minari.dataset.minari_storage import MinariStorage
from minari.namespace import create_namespace, list_local_namespaces
from minari.utils import _generate_dataset_metadata, _generate_dataset_path


# H5Py supports ints up to uint64
AUTOSEED_BIT_SIZE = 64


class DataCollector(gym.Wrapper):
    r"""Gymnasium environment wrapper that collects step data.

    This wrapper is meant to work as a temporary buffer of the environment data before creating a Minari dataset. The creation of the buffers
    that will be convert to a Minari dataset is agnostic to the user:

    .. code::

        import minari
        import gymnasium as gym

        env = minari.DataCollector(gym.make('EnvID'))

        env.reset()

        for _ in range(num_steps):
            action = env.action_space.sample()
            obs, rew, terminated, truncated, info = env.step()

            if terminated or truncated:
                env.reset()

        dataset = env.create_dataset(dataset_id="env_name/dataset_name-v(version)", **kwargs)

    Some of the characteristics of this wrapper:

        * The step data is stored per episode in dictionaries. This dictionaries are then stored in-memory in a global list buffer. The
          episode dictionaries contain items with list buffers as values for the main episode step datasets `observations`, `actions`,
          `terminations`, and `truncations`, the `infos` key can be a list or another nested dictionary with extra datasets.

        * A new episode dictionary buffer is created if the env.step(action) call returns `truncated` or `terminated`, or if the environment calls
          env.reset(). If calling reset and the previous episode was not `truncated` or `terminated`, this will automatically be `truncated`.

    """

    def __init__(
        self,
        env: gym.Env,
        step_data_callback: Type[StepDataCallback] = StepDataCallback,
        episode_metadata_callback: Type[
            EpisodeMetadataCallback
        ] = EpisodeMetadataCallback,
        record_infos: bool = False,
        observation_space: Optional[gym.Space] = None,
        action_space: Optional[gym.Space] = None,
        data_format: Optional[str] = None,
    ):
        """Initialize the data collector attributes and create the temporary directory for caching.

        Args:
            env (gym.Env): Gymnasium environment
            step_data_callback (type[StepDataCallback], optional): Callback class to edit/update step databefore storing to buffer. Defaults to StepDataCallback.
            episode_metadata_callback (type[EpisodeMetadataCallback], optional): Callback class to add custom metadata to episode group in HDF5 file. Defaults to EpisodeMetadataCallback.
            record_infos (bool, optional): If True record the info return key of each step. Defaults to False.
            observation_space (gym.Space): Observation space of the dataset. The default value is the environment observation space.
            action_space (gym.Space): Action space of the dataset. The default value is the environment action space.
            data_format (str, optional): Data format to store the data in the Minari dataset. If None (defaults), it will use the default format of MinariStorage.
        """
        super().__init__(env)
        self._step_data_callback = step_data_callback()
        self._episode_metadata_callback = episode_metadata_callback()

        self.datasets_path = os.environ.get("MINARI_DATASETS_PATH")
        if self.datasets_path is None:
            self.datasets_path = os.path.join(
                os.path.expanduser("~"), ".minari", "datasets"
            )
        if not os.path.exists(self.datasets_path):
            os.makedirs(self.datasets_path)
        self.data_format = data_format

        if observation_space is None:
            observation_space = env.observation_space
        self._observation_space = observation_space
        if action_space is None:
            action_space = env.action_space
        self._action_space = action_space

        self._record_infos = record_infos
        self._buffer: Optional[EpisodeBuffer] = None
        self._episode_id = 0
        self._reset_storage()

    def _reset_storage(self):
        self._episode_id = 0
        self._tmp_dir = tempfile.TemporaryDirectory(dir=self.datasets_path)
        data_format_kwarg = (
            {"data_format": self.data_format} if self.data_format is not None else {}
        )
        self._storage = MinariStorage.new(
            self._tmp_dir.name,
            observation_space=self._observation_space,
            action_space=self._action_space,
            env_spec=self.env.spec,
            **data_format_kwarg,
        )

    def step(
        self, action: ActType
    ) -> tuple[ObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        """Gymnasium step method."""
        obs, rew, terminated, truncated, info = self.env.step(action)

        step_data = self._step_data_callback(
            env=self.env,
            obs=obs,
            info=info,
            action=action,
            rew=rew,
            terminated=terminated,
            truncated=truncated,
        )

        if not self._storage.observation_space.contains(step_data["observation"]):
            logging.warning(
                "Observation is not in observation space.\n"
                f"Observation: {step_data['observation']}\nSpace: {self._storage.observation_space}"
            )
        if not self._storage.action_space.contains(step_data["action"]):
            logging.warning(
                "Action is not in action space.\n"
                f"Action: {step_data['action']}\nSpace: {self._storage.action_space}",
            )

        assert self._buffer is not None
        if not self._record_infos:
            step_data["info"] = {}
        self._buffer = self._buffer.add_step_data(step_data)

        if step_data["termination"] or step_data["truncation"]:
            self._storage.update_episodes([self._buffer])
            self._episode_id += 1
            self._buffer = EpisodeBuffer(
                id=self._episode_id,
                observations=step_data["observation"],
                infos=step_data["info"],
            )

        return obs, rew, terminated, truncated, info

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict[str, Any] | None = None,
    ) -> tuple[ObsType, dict[str, Any]]:
        """Gymnasium environment reset.

        If no seed is set, one will be automatically generated, for reproducibility,
        unless ``minari_autoseed=False`` in the ``options`` dictionary.

        Args:
            seed (optional int): The seed that is used to initialize the environment's PRNG.
                If no seed is specified, one will be automatically generated (by default).
            options (optional dict): Additional information to specify how the environment is reset.
                Set ``minari_autoseed=False`` to disable automatic seeding.

        Returns:
            observation (ObsType): Observation of the initial state.
            info (dictionary): Auxiliary information complementing ``observation``.
        """
        self._flush_to_storage()

        autoseed_enabled = (not options) or options.get("minari_autoseed", True)
        if seed is None and autoseed_enabled:
            seed = secrets.randbits(AUTOSEED_BIT_SIZE)

        obs, info = self.env.reset(seed=seed, options=options)
        step_data = self._step_data_callback(env=self.env, obs=obs, info=info)

        if not self._storage.observation_space.contains(step_data["observation"]):
            logging.warning(
                "Observation is not in observation space.\n"
                f"Observation: {step_data['observation']}\nSpace: {self._storage.observation_space}"
            )

        self._buffer = EpisodeBuffer(
            id=self._episode_id,
            seed=seed,
            options=options,
            observations=step_data["observation"],
            infos=step_data["info"] if self._record_infos else None,
        )
        return obs, info

    def add_to_dataset(self, dataset: MinariDataset):
        """Add extra data to Minari dataset from collector environment buffers (DataCollector).

        Args:
            dataset (MinariDataset): Dataset to add the data
        """
        self._flush_to_storage()

        first_id = dataset.storage.total_episodes
        dataset.storage.update_from_storage(self._storage)
        if dataset.episode_indices is not None:
            new_ids = first_id + np.arange(self._storage.total_episodes)
            dataset.episode_indices = np.append(dataset.episode_indices, new_ids)

        self._reset_storage()

    def create_dataset(
        self,
        dataset_id: str,
        eval_env: Optional[str | gym.Env | EnvSpec] = None,
        algorithm_name: Optional[str] = None,
        author: Optional[str | set] = None,
        author_email: Optional[str | set] = None,
        code_permalink: Optional[str] = None,
        ref_min_score: Optional[float] = None,
        ref_max_score: Optional[float] = None,
        expert_policy: Optional[Callable[[ObsType], ActType]] = None,
        num_episodes_average_score: int = 100,
        description: Optional[str] = None,
        requirements: Optional[list] = None,
    ):
        """Create a Minari dataset using the data collected from stepping with a Gymnasium environment wrapped with a `DataCollector` Minari wrapper.

        The ``dataset_id`` parameter corresponds to the name of the dataset, with the syntax as follows:
        ``(namespace/)(env_name/)dataset_name(-v[version])`` where ``env_name`` identifies the name of the environment used to generate the dataset. The `namespace` is optional.
        This ``dataset_id`` is used to load the Minari datasets with :meth:`minari.load_dataset`.

        Args:
            dataset_id (str): name id to identify Minari dataset
            eval_env (str | gym.Env | EnvSpec, optional): Gymnasium environment(gym.Env)/environment id(str)/environment spec(EnvSpec) to use for evaluation with the dataset. After loading the dataset, the environment can be recovered as follows: `MinariDataset.recover_environment(eval_env=True).
                                                    If None the `env` used to collect the buffer data should be used for evaluation.
            algorithm_name (str, optional): name of the algorithm used to collect the data. Defaults to None.
            author (str | set, optional): name of the author(s) that generated the dataset. Defaults to None.
            author_email (str | set, optional): email(s) of the author(s) that generated the dataset. Defaults to None.
            code_permalink (str, optional): link to relevant code used to generate the dataset. Defaults to None.
            ref_min_score(float, optional): minimum reference score from the average returns of a random policy. This value is later used to normalize a score with :meth:`minari.get_normalized_score`. If default None the value will be estimated with a default random policy.
            ref_max_score (float, optional): maximum reference score from the average returns of a hypothetical expert policy. This value is used in :meth:`minari.get_normalized_score`. Default None.
            expert_policy (Callable[[ObsType], ActType], optional): policy to compute `ref_max_score` by averaging the returns over a number of episodes equal to  `num_episodes_average_score`.
                                                                            `ref_max_score` and `expert_policy` can't be passed at the same time. Default to None
            num_episodes_average_score (int): number of episodes to average over the returns to compute `ref_min_score` and `ref_max_score`. Default to 100.
            description (str, optional): description of the dataset being created. Defaults to None.
            requirements (list of str, optional): list of requirements in pip-style to load the environment and reproduce the dataset. For example, `mujoco>=3.1.0,<3.2.0`, which indicate the supported version range for mujoco package. Defaults to None.

        Returns:
            MinariDataset
        """
        namespace = parse_dataset_id(dataset_id)[0]

        if namespace is not None and namespace not in list_local_namespaces():
            create_namespace(namespace)

        dataset_path = _generate_dataset_path(dataset_id)
        metadata: Dict[str, Any] = _generate_dataset_metadata(
            dataset_id,
            copy.deepcopy(self.env.spec),
            eval_env,
            algorithm_name,
            author,
            author_email,
            code_permalink,
            ref_min_score,
            ref_max_score,
            expert_policy,
            num_episodes_average_score,
            description,
            requirements,
        )

        self._save_to_disk(dataset_path, metadata)

        dataset = MinariDataset(dataset_path)
        dataset.storage.update_metadata({"dataset_size": dataset.storage.get_size()})
        return dataset

    def _flush_to_storage(self):
        if self._buffer is not None and len(self._buffer) > 0:
            if not self._buffer.terminations[-1]:
                self._buffer.truncations[-1] = True
            self._storage.update_episodes([self._buffer])
            self._episode_id += 1
        self._buffer = None

    def _save_to_disk(
        self, path: str | os.PathLike, dataset_metadata: Dict[str, Any] = {}
    ):
        """Save all in-memory buffer data and move temporary files to a permanent location in disk.

        Args:
            path (str): path to store the dataset, e.g.: '/home/foo/datasets/data'
            dataset_metadata (Dict, optional): additional metadata to add to the dataset file. Defaults to {}.
        """
        self._flush_to_storage()

        assert (
            "observation_space" not in dataset_metadata.keys()
        ), "'observation_space' is not allowed as an optional key."
        assert (
            "action_space" not in dataset_metadata.keys()
        ), "'action_space' is not allowed as an optional key."
        assert (
            "env_spec" not in dataset_metadata.keys()
        ), "'env_spec' is not allowed as an optional key."
        self._storage.update_metadata(dataset_metadata)

        episode_metadata = self._storage.apply(self._episode_metadata_callback)
        self._storage.update_episode_metadata(episode_metadata)

        files = os.listdir(self._storage.data_path)
        for file in files:
            shutil.move(
                os.path.join(self._storage.data_path, file),
                os.path.join(path, file),
            )

        self._reset_storage()

    def close(self):
        """Close the DataCollector.

        Clear buffer and close temporary directory.
        """
        super().close()
        self._buffer = None
        shutil.rmtree(self._tmp_dir.name)


def _check_infos_same_shape(info_1: dict, info_2: dict):
    if info_1.keys() != info_2.keys():
        return False
    for key in info_1.keys():
        if type(info_1[key]) is not type(info_2[key]):
            return False
        if isinstance(info_1[key], dict):
            return _check_infos_same_shape(info_1[key], info_2[key])
        elif isinstance(info_1[key], np.ndarray):
            return (info_1[key].shape == info_2[key].shape) and (
                info_1[key].dtype == info_2[key].dtype
            )
    return True



import pytest
from typer.testing import CliRunner

from minari.cli import app
from minari.storage.hosting import list_remote_datasets
from tests.dataset.test_dataset_download import get_latest_compatible_dataset_id


runner = CliRunner()


def test_list_app():
    result = runner.invoke(app, ["list", "local", "--all"])
    assert result.exit_code == 0
    # some of the other columns may be cut off by Rich
    assert "Name" in result.stdout

    result = runner.invoke(app, ["list", "remote"])
    assert result.exit_code == 0
    assert "Minari datasets in Farama server" in result.stdout


@pytest.mark.parametrize(
    "dataset_id",
    [get_latest_compatible_dataset_id(namespace="D4RL/pen", dataset_name="human")],
)
def test_dataset_download_then_delete(dataset_id: str):
    """Test download dataset invocation from CLI.

    The downloading functionality is already tested in test_dataset_download.py so this
    is primarily to assert that the CLI is working as expected.
    """
    result = runner.invoke(app, ["download", dataset_id])

    assert result.exit_code == 0
    assert f"Downloading {dataset_id} from Farama servers..." in result.stdout
    assert f"Dataset {dataset_id} downloaded to" in result.stdout

    result = runner.invoke(app, ["delete", dataset_id], input="n")
    assert result.exit_code == 1  # aborted but no error
    assert "Aborted" in result.stdout

    result = runner.invoke(app, ["delete", dataset_id], input="😳")
    assert result.exit_code == 1
    assert "Error: invalid input" in result.stdout

    result = runner.invoke(app, ["delete", dataset_id], input="y")
    assert result.exit_code == 0
    assert f"Dataset {dataset_id} deleted!" in result.stdout


@pytest.mark.parametrize(
    "dataset_id",
    list_remote_datasets(compatible_minari_version=True),
)
def test_minari_show(dataset_id):
    result = runner.invoke(app, ["show", dataset_id])
    assert result.exit_code == 0



import gymnasium as gym
import pytest

import minari
from minari.data_collector.data_collector import DataCollector
from minari.dataset.minari_dataset import MinariDataset
from minari.namespace import (
    create_namespace,
    delete_namespace,
    download_namespace_metadata,
    get_namespace_metadata,
    list_local_namespaces,
    list_remote_namespaces,
    update_namespace_metadata,
)
from tests.common import (
    check_data_integrity,
    check_load_and_delete_dataset,
    create_dummy_dataset_with_collecter_env_helper,
    get_latest_compatible_dataset_id,
)


@pytest.mark.parametrize(
    "namespace",
    [
        "example_-123",
        "nested/nested/nested/namespace",
    ],
)
@pytest.mark.parametrize(
    "description, metadata, combined_metadata",
    [
        (None, {}, {}),
        ("my_desc", {}, {"description": "my_desc"}),
        (None, {"key": [1]}, {"key": [1]}),
        ("my_desc", {"key": [1]}, {"description": "my_desc", "key": [1]}),
    ],
)
def test_create_namespace(namespace, description, metadata, combined_metadata):
    create_namespace(namespace, description, **metadata)
    assert namespace in list_local_namespaces()
    assert get_namespace_metadata(namespace) == combined_metadata
    delete_namespace(namespace)


@pytest.mark.parametrize("namespace", ["test_namespace"])
def test_namespace_update(namespace):
    create_namespace(namespace)

    with pytest.raises(ValueError, match="Namespace 'test_namespace' already exists"):
        create_namespace(namespace)

    update_namespace_metadata(namespace, description="a new definition")
    assert get_namespace_metadata(namespace) == {"description": "a new definition"}


@pytest.mark.parametrize("namespace", ["test_namespace"])
def test_create_nested_namespaces(namespace):
    create_namespace(namespace, description="my description")
    assert list_local_namespaces() == [namespace]
    assert get_namespace_metadata(namespace) == {"description": "my description"}

    nested_namespace = f"{namespace}/nested"
    create_namespace(nested_namespace, description="is nested")
    assert list_local_namespaces() == [namespace, nested_namespace]
    assert get_namespace_metadata(nested_namespace) == {"description": "is nested"}


def test_nonexistent_namespaces():
    with pytest.raises(ValueError, match="does not exist"):
        update_namespace_metadata("does_not/exist")

    with pytest.raises(ValueError, match="does not exist"):
        get_namespace_metadata("does_not/exist")


@pytest.mark.parametrize("namespace", ["/", "./", "../", "/namespace", "namespace/"])
def test_create_invalid_namespace(namespace):
    with pytest.raises(ValueError, match="Malformed namespace"):
        create_namespace(namespace)


@pytest.mark.parametrize("namespace", ["nested/namespace"])
def test_create_namespaced_datasets(namespace):
    env = gym.make("CartPole-v1")
    env = DataCollector(env)

    dataset_id_1 = f"{namespace}/test-v1"
    create_dummy_dataset_with_collecter_env_helper(dataset_id_1, env)
    assert list_local_namespaces() == [namespace]
    assert get_namespace_metadata(namespace) == {}

    update_namespace_metadata(namespace, description="new description")
    assert get_namespace_metadata(namespace) == {"description": "new description"}

    # Creating a new dataset in the same namespace doesn't change the namespace metadata
    dataset_id_2 = f"{namespace}/test-v2"
    create_dummy_dataset_with_collecter_env_helper(dataset_id_2, env)
    assert list_local_namespaces() == [namespace]
    assert get_namespace_metadata(namespace) == {"description": "new description"}

    assert list(minari.list_local_datasets().keys()) == [dataset_id_1, dataset_id_2]

    # Check that can only delete when empty
    with pytest.raises(ValueError, match="is not empty"):
        delete_namespace(namespace)

    check_load_and_delete_dataset(dataset_id_1)
    check_load_and_delete_dataset(dataset_id_2)
    delete_namespace(namespace)


def test_download_namespace_dataset():
    namespace = "D4RL/kitchen"
    kitchen_complete = get_latest_compatible_dataset_id(namespace, "complete")
    kitchen_mix = get_latest_compatible_dataset_id(namespace, "mixed")

    remote_datasets = minari.list_remote_datasets()
    assert kitchen_complete in remote_datasets

    minari.download_dataset(kitchen_complete)
    assert set(minari.list_local_datasets().keys()) == {kitchen_complete}
    assert list_local_namespaces() == [namespace]
    assert get_namespace_metadata(namespace) == {}

    minari.download_dataset(kitchen_mix)
    assert set(minari.list_local_datasets().keys()) == {kitchen_complete, kitchen_mix}
    assert list_local_namespaces() == [namespace]
    assert get_namespace_metadata(namespace) == {}

    with pytest.warns(UserWarning, match="Skipping Download."):
        minari.download_dataset(kitchen_complete)

    dataset = minari.load_dataset(kitchen_complete)
    assert isinstance(dataset, MinariDataset)

    check_data_integrity(dataset, list(dataset.episode_indices))

    minari.delete_dataset(kitchen_complete)
    assert set(minari.list_local_datasets().keys()) == {kitchen_mix}
    assert list_local_namespaces() == [namespace]


@pytest.mark.parametrize("namespace", ["D4RL/door", "D4RL/pen"])
def test_download_namespace_metadata(namespace):
    assert len(list_local_namespaces()) == 0
    assert namespace in list_remote_namespaces()

    download_namespace_metadata(namespace)
    assert list_local_namespaces() == [namespace]
    get_namespace_metadata(namespace)
    delete_namespace(namespace)

    metadata = {"description": "Conflicting description"}
    create_namespace(namespace, **metadata)
    with pytest.warns(UserWarning, match="Skipping update for namespace"):
        download_namespace_metadata(namespace)

    assert get_namespace_metadata(namespace) == metadata

    download_namespace_metadata(namespace, overwrite=True)
    assert get_namespace_metadata(namespace) != metadata

    with pytest.raises(ValueError, match="doesn't exist in the remote Farama server."):
        download_namespace_metadata("non_existent_namespace")



from __future__ import annotations

import copy
import importlib.metadata
import os
import re
import warnings
from typing import Any, Callable, Dict, Iterable, List, Optional

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType
from gymnasium.envs.registration import EnvSpec
from gymnasium.error import NameNotFound
from gymnasium.wrappers import RecordEpisodeStatistics  # type: ignore

from minari.data_collector.episode_buffer import EpisodeBuffer
from minari.dataset.minari_dataset import MinariDataset
from minari.dataset.minari_storage import MinariStorage
from minari.serialization import deserialize_space
from minari.storage.datasets_root_dir import get_dataset_path


# Use importlib due to circular import when: "from minari import __version__"
__version__ = importlib.metadata.version("minari")


def validate_datasets_to_combine(
    datasets_to_combine: List[MinariDataset],
) -> EnvSpec | None:
    """Check if the given datasets can be combined.

    Tests if the datasets were created with the same environment (`env_spec`) and re-calculates the
    `max_episode_steps` argument.

    Also checks that the datasets obs/act spaces are the same.

    Args:
        datasets_to_combine (List[MinariDataset]): list of MinariDataset to combine

    Returns:
        combined_dataset_env_spec (EnvSpec): the resulting EnvSpec of combining the MinariDatasets

    """
    # get first among the dataset's env_spec which is not None
    first_not_none_env_spec = next(
        (
            dataset.spec.env_spec
            for dataset in datasets_to_combine
            if dataset.spec.env_spec is not None
        ),
        None,
    )

    # early return where all datasets have no env_spec
    if first_not_none_env_spec is None:
        return None

    common_env_spec = copy.deepcopy(first_not_none_env_spec)

    # updating the common_env_spec's max_episode_steps & checking equivalence of all env specs
    for dataset in datasets_to_combine:
        assert isinstance(dataset, MinariDataset)
        env_spec = dataset.spec.env_spec
        if env_spec is not None:
            if (
                common_env_spec.max_episode_steps is None
                or env_spec.max_episode_steps is None
            ):
                common_env_spec.max_episode_steps = None
            else:
                common_env_spec.max_episode_steps = max(
                    common_env_spec.max_episode_steps, env_spec.max_episode_steps
                )
            # setting max_episode_steps in object's copy to same value for sake of checking equality
            env_spec_copy = copy.deepcopy(env_spec)
            env_spec_copy.max_episode_steps = common_env_spec.max_episode_steps
            if env_spec_copy != common_env_spec:
                raise ValueError(
                    "The datasets to be combined have different values for `env_spec` attribute."
                )
        else:
            raise ValueError(
                "Cannot combine datasets having env_spec with those having no env_spec."
            )

    return common_env_spec


class RandomPolicy:
    """A random action selection policy to compute `ref_min_score`."""

    def __init__(self, env: gym.Env):
        self.action_space = env.action_space
        self.action_space.seed(123)
        self.observation_space = env.observation_space

    def __call__(self, observation: ObsType) -> ActType:
        assert self.observation_space.contains(observation)
        return self.action_space.sample()


def combine_datasets(datasets_to_combine: List[MinariDataset], new_dataset_id: str):
    """Combine a group of MinariDataset in to a single dataset with its own name id.

    The new dataset will contain a metadata attribute `combined_datasets` containing a list
    with the dataset names that were combined to form this new Minari dataset.

    Args:
        datasets_to_combine (list[MinariDataset]): list of datasets to be combined
        new_dataset_id (str): name id for the newly created dataset

    Returns:
        combined_dataset (MinariDataset): the resulting MinariDataset
    """
    combined_dataset_env_spec = validate_datasets_to_combine(datasets_to_combine)

    new_dataset_path = get_dataset_path(new_dataset_id)
    new_dataset_path.mkdir()
    new_storage = MinariStorage.new(
        new_dataset_path.joinpath("data"),
        env_spec=combined_dataset_env_spec,
        observation_space=datasets_to_combine[0].observation_space,
        action_space=datasets_to_combine[0].action_space,
        data_format=datasets_to_combine[0].storage.FORMAT,
    )

    new_storage.update_metadata(
        {
            "dataset_id": new_dataset_id,
            "combined_datasets": [
                dataset.spec.dataset_id for dataset in datasets_to_combine
            ],
            "minari_version": __version__,
        }
    )

    for dataset in datasets_to_combine:
        new_storage.update_from_storage(dataset.storage)

    return MinariDataset(new_storage)


def split_dataset(
    dataset: MinariDataset, sizes: List[int], seed: Optional[int] = None
) -> List[MinariDataset]:
    """Split a MinariDataset in multiple datasets.

    Args:
        dataset (MinariDataset): the MinariDataset to split
        sizes (List[int]): sizes of the resulting datasets
        seed (Optional[int]): random seed

    Returns:
        datasets (List[MinariDataset]): resulting list of datasets
    """
    if sum(sizes) > dataset.total_episodes:
        raise ValueError(
            "Incompatible arguments: the sum of sizes exceeds ",
            f"the number of episodes in the dataset ({dataset.total_episodes})",
        )
    generator = np.random.default_rng(seed=seed)
    indices = generator.permutation(dataset.episode_indices)
    out_datasets = []
    start_idx = 0
    for length in sizes:
        end_idx = start_idx + length
        slice_dataset = MinariDataset(
            dataset.spec.data_path, indices[start_idx:end_idx]
        )
        out_datasets.append(slice_dataset)
        start_idx = end_idx

    return out_datasets


def get_average_reference_score(
    env: gym.Env,
    policy: Callable[[ObsType], ActType],
    num_episodes: int,
) -> float:
    env = RecordEpisodeStatistics(env, num_episodes)
    episode_returns = []
    obs, _ = env.reset(seed=123)
    for _ in range(num_episodes):
        while True:
            action = policy(obs)
            obs, _, terminated, truncated, info = env.step(
                action  # pyright: ignore[reportGeneralTypeIssues]
            )
            if terminated or truncated:
                episode_returns.append(info["episode"]["r"])
                obs, _ = env.reset()
                break

    mean_ref_score = np.mean(episode_returns, dtype=np.float32)
    return float(mean_ref_score)


def _generate_dataset_path(dataset_id: str) -> str | os.PathLike:
    """Checks if the dataset already exists locally, then create and return the data storage directory."""
    dataset_path = get_dataset_path(dataset_id)
    if os.path.exists(dataset_path):
        raise ValueError(
            f"A Minari dataset with ID {dataset_id} already exists and it cannot be overridden. Please use a different dataset name or version."
        )

    dataset_path = os.path.join(dataset_path, "data")
    os.makedirs(dataset_path)

    return dataset_path


def _generate_dataset_metadata(
    dataset_id: str,
    env_spec: Optional[EnvSpec],
    eval_env: Optional[str | gym.Env | EnvSpec],
    algorithm_name: Optional[str],
    author: Optional[str | set[str]],
    author_email: Optional[str | set[str]],
    code_permalink: Optional[str],
    ref_min_score: Optional[float],
    ref_max_score: Optional[float],
    expert_policy: Optional[Callable[[ObsType], ActType]],
    num_episodes_average_score: int,
    description: Optional[str],
    requirements: Optional[list],
) -> Dict[str, Any]:
    """Return the metadata dictionary of the dataset."""
    dataset_metadata: Dict[str, Any] = {
        "dataset_id": dataset_id,
    }
    # NoneType warnings
    if code_permalink is None:
        warnings.warn(
            "`code_permalink` is set to None. For reproducibility purposes it is highly recommended to link your dataset to versioned code.",
            UserWarning,
        )
    else:
        dataset_metadata["code_permalink"] = code_permalink

    if author is None:
        warnings.warn(
            "`author` is set to None. For longevity purposes it is highly recommended to provide an author name.",
            UserWarning,
        )
    else:
        dataset_metadata["author"] = {author} if isinstance(author, str) else author

    if author_email is None:
        warnings.warn(
            "`author_email` is set to None. For longevity purposes it is highly recommended to provide an author email, or some other obvious contact information.",
            UserWarning,
        )
    else:
        dataset_metadata["author_email"] = (
            {author_email} if isinstance(author_email, str) else author_email
        )

    if algorithm_name is None:
        warnings.warn(
            "`algorithm_name` is set to None. For reproducibility purpose it's highly recommended to set your algorithm",
            UserWarning,
        )
    else:
        dataset_metadata["algorithm_name"] = algorithm_name

    if description is None:
        warnings.warn(
            "`description` is set to None. For longevity purposes it is highly recommended to provide a description of the dataset",
            UserWarning,
        )
    else:
        dataset_metadata["description"] = description

    dataset_metadata["minari_version"] = __version__

    if expert_policy is not None and ref_max_score is not None:
        raise ValueError(
            "Can't pass a value for `expert_policy` and `ref_max_score` at the same time."
        )

    if eval_env is None:
        warnings.warn(
            f"`eval_env` is set to None. If another environment is intended to be used for evaluation please specify corresponding Gymnasium environment (gym.Env | gym.envs.registration.EnvSpec). "
            f"If None the environment used to collect the data (`env={env_spec}`) will be used for this purpose.",
            UserWarning,
        )
        eval_env_spec = env_spec
    else:
        if isinstance(eval_env, str):
            eval_env_spec = gym.spec(eval_env)
        elif isinstance(eval_env, EnvSpec):
            eval_env_spec = eval_env
        elif isinstance(eval_env, gym.Env):
            eval_env_spec = eval_env.spec
        else:
            raise ValueError(
                "The `eval_env` argument must be of types str|EnvSpec|gym.Env"
            )
        assert eval_env_spec is not None
        dataset_metadata["eval_env_spec"] = eval_env_spec.to_json()

    if env_spec is None:
        warnings.warn(
            "env_spec is None, no environment spec is provided during collection for this dataset",
            UserWarning,
        )

    if eval_env_spec is not None and (
        expert_policy is not None or ref_max_score is not None
    ):
        env_ref_score = gym.make(eval_env_spec)
        if ref_min_score is None:
            ref_min_score = get_average_reference_score(
                env_ref_score, RandomPolicy(env_ref_score), num_episodes_average_score
            )

        if expert_policy is not None:
            ref_max_score = get_average_reference_score(
                env_ref_score, expert_policy, num_episodes_average_score
            )
        dataset_metadata["ref_max_score"] = ref_max_score
        dataset_metadata["ref_min_score"] = ref_min_score
        dataset_metadata["num_episodes_average_score"] = num_episodes_average_score

    if requirements is not None:
        dataset_metadata["requirements"] = requirements
    return dataset_metadata


def create_dataset_from_buffers(
    dataset_id: str,
    buffer: List[EpisodeBuffer],
    env: Optional[str | gym.Env | EnvSpec] = None,
    eval_env: Optional[str | gym.Env | EnvSpec] = None,
    algorithm_name: Optional[str] = None,
    author: Optional[str | set[str]] = None,
    author_email: Optional[str | set[str]] = None,
    code_permalink: Optional[str] = None,
    action_space: Optional[gym.spaces.Space] = None,
    observation_space: Optional[gym.spaces.Space] = None,
    ref_min_score: Optional[float] = None,
    ref_max_score: Optional[float] = None,
    expert_policy: Optional[Callable[[ObsType], ActType]] = None,
    num_episodes_average_score: int = 100,
    description: Optional[str] = None,
    data_format: Optional[str] = None,
    requirements: Optional[list] = None,
):
    """Create Minari dataset from a list of episode dictionary buffers.

    The ``dataset_id`` parameter corresponds to the name of the dataset, with the syntax as follows:
    ``(namespace/)(env_name/)dataset_name(-v[version])`` where ``env_name`` identifies, if present, the name of the environment used to generate the dataset ``dataset_name`` and ``namespace`` optionally groups datasets together.
    This ``dataset_id`` is used to load the Minari datasets with :meth:`minari.load_dataset`.

    Args:
        dataset_id (str): name id to identify Minari dataset.
        buffer (list[EpisodeBuffer]): list of episode buffer with data.
        env (str | gym.Env | EnvSpec, optional): Gymnasium environment(gym.Env)/environment id(str)/environment spec(EnvSpec) used to collect the buffer data. Defaults to None.
        eval_env (str | gym.Env | EnvSpec, optional): Gymnasium environment(gym.Env)/environment id(str)/environment spec(EnvSpec) to use for evaluation with the dataset. After loading the dataset, the environment can be recovered as follows: `MinariDataset.recover_environment(eval_env=True).
                                                If None, and if the `env` used to collect the buffer data is available, latter will be used for evaluation.
        algorithm_name (str, optional): name of the algorithm used to collect the data. Defaults to None.
        author (str | set, optional): name of the author(s) that generated the dataset. Defaults to None.
        author_email (str | set, optional): email(s) of the author(s) that generated the dataset. Defaults to None.
        code_permalink (str, optional): link to relevant code used to generate the dataset. Defaults to None.
        ref_min_score (float, optional): minimum reference score from the average returns of a random policy. This value is later used to normalize a score with :meth:`minari.get_normalized_score`. If default None the value will be estimated with a default random policy.
                                                    Also note that this attribute will be added to the Minari dataset only if `ref_max_score` or `expert_policy` are assigned a valid value other than None.
        ref_max_score (float, optional): maximum reference score from the average returns of a hypothetical expert policy. This value is used in `MinariDataset.get_normalized_score()`. Default None.
        expert_policy (Callable[[ObsType], ActType], optional): policy to compute `ref_max_score` by averaging the returns over a number of episodes equal to  `num_episodes_average_score`.
                                                                        `ref_max_score` and `expert_policy` can't be passed at the same time. Default to None
        num_episodes_average_score (int): number of episodes to average over the returns to compute `ref_min_score` and `ref_max_score`. Default to 100.
                observation_space:
        action_space (gym.spaces.Space, optional): action space of the environment. If None (default) use the environment action space.
        observation_space (gym.spaces.Space, optional): observation space of the environment. If None (default) use the environment observation space.
        description (str, optional): description of the dataset being created. Defaults to None.
        data_format (str, optional): Data format to store the data in the Minari dataset. If None (defaults), it will use the default format of MinariStorage.
        requirements (list of str, optional): list of requirements in pip-style to load the environment and reproduce the dataset. For example, `mujoco>=3.1.0,<3.2.0`, which indicate the supported version range for mujoco package. Defaults to None.

    Returns:
        MinariDataset
    """
    dataset_path = _generate_dataset_path(dataset_id)

    if isinstance(env, str):
        env_spec = gym.spec(env)
    elif isinstance(env, EnvSpec):
        env_spec = env
    elif isinstance(env, gym.Env):
        env_spec = env.spec
    elif env is None:
        if observation_space is None or action_space is None:
            raise ValueError(
                "Both observation space and action space must be provided, if env is None"
            )
        env_spec = None
    else:
        raise ValueError("The `env` argument must be of types str|EnvSpec|gym.Env|None")

    if isinstance(env, (str, EnvSpec)):
        env = gym.make(env)
    if observation_space is None:
        assert isinstance(env, gym.Env)
        observation_space = env.observation_space
    if action_space is None:
        assert isinstance(env, gym.Env)
        action_space = env.action_space

    metadata = _generate_dataset_metadata(
        dataset_id,
        env_spec,
        eval_env,
        algorithm_name,
        author,
        author_email,
        code_permalink,
        ref_min_score,
        ref_max_score,
        expert_policy,
        num_episodes_average_score,
        description,
        requirements,
    )

    data_format_kwarg = {"data_format": data_format} if data_format is not None else {}
    storage = MinariStorage.new(
        dataset_path,
        observation_space=observation_space,
        action_space=action_space,
        env_spec=env_spec,
        **data_format_kwarg,
    )

    storage.update_metadata(metadata)
    storage.update_episodes(buffer)
    storage.update_metadata({"dataset_size": storage.get_size()})

    return MinariDataset(storage)


def get_normalized_score(dataset: MinariDataset, returns: np.ndarray) -> np.ndarray:
    r"""Normalize undiscounted return of an episode.

    This function was originally provided in the `D4RL repository <https://github.com/Farama-Foundation/D4RL/blob/71a9549f2091accff93eeff68f1f3ab2c0e0a288/d4rl/offline_env.py#L71>`_.
    The computed normalized episode return (normalized score) facilitates the comparison of algorithm performance across different tasks. The returned normalized score will be in a range between 0 and 1.
    Where 0 corresponds to the minimum reference score calculated as the average of episode returns collected from a random policy in the environment, and 1 corresponds to a maximum reference score computed as
    the average of episode returns from an hypothetical expert policy. These two values are stored as optional attributes in a MinariDataset as `ref_min_score` and `ref_max_score` respectively.

    The formula to normalize an episode return is:

    .. math:: normalize\_score = \frac{return - ref\_min\_score}{ref\_max\_score - ref\_min\_score}

    .. warning:: This utility function is under testing and will not be available in every Minari dataset. For now, only the datasets imported from D4RL will contain the `ref_min_score` and `ref_max_score` attributes.

    Args:
        dataset (MinariDataset): the MinariDataset with respect to which normalize the score. Must contain the reference score attributes `ref_min_score` and `ref_max_score`.
        returns (np.ndarray): a single value or array of episode undiscounted returns to normalize.

    Returns:
        normalized_scores
    """
    ref_min_score = dataset.storage.metadata.get("ref_min_score")
    ref_max_score = dataset.storage.metadata.get("ref_max_score")

    if ref_min_score is None or ref_max_score is None:
        raise ValueError(
            f"Reference score not provided for dataset {dataset.spec.dataset_id}. Can't compute the normalized score."
        )

    return (returns - ref_min_score) / (ref_max_score - ref_min_score)


def get_env_spec_dict(env_spec: EnvSpec) -> Dict[str, str]:
    """Create dict of the environment specs, including observation and action space."""
    try:
        env = gym.make(env_spec.id)
        action_space_table = env.action_space.__repr__().replace("\n", "")
        observation_space_table = env.observation_space.__repr__().replace("\n", "")
    except NameNotFound:
        action_space_table, observation_space_table = None, None

    md_dict = {"ID": env_spec.id}
    if observation_space_table is not None:
        md_dict["Observation Space"] = f"`{re.sub(' +', ' ', observation_space_table)}`"
    if action_space_table is not None:
        md_dict["Action Space"] = f"`{re.sub(' +', ' ', action_space_table)}`"

    md_dict.update(
        {
            "entry_point": f"`{env_spec.entry_point}`",
            "max_episode_steps": str(env_spec.max_episode_steps),
            "reward_threshold": str(env_spec.reward_threshold),
            "nondeterministic": f"`{env_spec.nondeterministic}`",
            "order_enforce": f"`{env_spec.order_enforce}`",
            "disable_env_checker": f"`{env_spec.disable_env_checker}`",
            "kwargs": f"`{env_spec.kwargs}`",
            "additional_wrappers": f"`{env_spec.additional_wrappers}`",
            "vector_entry_point": f"`{env_spec.vector_entry_point}`",
        }
    )

    return {k: str(v) for k, v in md_dict.items()}


def get_dataset_spec_dict(dataset_spec: Dict) -> Dict[str, str]:
    """Create dict of the dataset specs, including observation and action space."""
    code_link = dataset_spec["code_permalink"]
    action_space = dataset_spec.get("action_space")
    obs_space = dataset_spec.get("observation_space")

    md_dict = {
        "Total Steps": str(dataset_spec["total_steps"]),
        "Total Episodes": str(dataset_spec["total_episodes"]),
    }

    if obs_space is not None:
        if isinstance(obs_space, (dict, str)):
            obs_space = deserialize_space(obs_space)
        dataset_observation_space = obs_space.__repr__().replace("\n", "")
        md_dict["Dataset Observation Space"] = f"`{dataset_observation_space}`"

    if action_space is not None:
        if isinstance(action_space, (dict, str)):
            action_space = deserialize_space(action_space)
        dataset_action_space = action_space.__repr__().replace("\n", "")
        md_dict["Dataset Action Space"] = f"`{dataset_action_space}`"

    from minari import supported_dataset_versions

    version = dataset_spec["minari_version"]
    supported = (
        "supported" if version in supported_dataset_versions else "not supported"
    )
    author = dataset_spec.get("author", "Not provided")
    if isinstance(author, Iterable):
        author = ", ".join(author)
    email = dataset_spec.get("author_email", "Not provided")
    if isinstance(email, Iterable):
        email = ", ".join(email)
    assert isinstance(author, str)
    assert isinstance(email, str)
    md_dict.update(
        {
            "Algorithm": dataset_spec.get("algorithm_name", "Not provided"),
            "Author": author,
            "Email": email,
            "Code Permalink": f"[{code_link}]({code_link})",
            "Minari Version": f"`{version}` ({supported})",
            "Download": f"`minari download {dataset_spec['dataset_id']}`",
        }
    )

    return md_dict



from typing import Callable, Dict, List, Type

from minari.dataset.minari_storage import MinariStorage


def get_arrow_storage() -> Type[MinariStorage]:
    from .arrow_storage import ArrowStorage

    return ArrowStorage


def get_hdf5_storage() -> Type[MinariStorage]:
    from .hdf5_storage import HDF5Storage

    return HDF5Storage


_registry: Dict[str, Callable[[], Type[MinariStorage]]] = {
    "arrow": get_arrow_storage,
    "hdf5": get_hdf5_storage,
}


def get_minari_storage(storage_type: str) -> Type[MinariStorage]:
    storage = _registry[storage_type]()
    assert (
        storage.FORMAT == storage_type
    ), f"Storage type mismatch: {storage.FORMAT} != {storage_type}"
    return storage


def get_storage_keys() -> List[str]:
    return list(_registry.keys())



from __future__ import annotations

import json
import pathlib
from itertools import zip_longest
from typing import Any, Dict, Iterable, Optional, Sequence

import gymnasium as gym
import numpy as np


try:
    import pyarrow as pa
    import pyarrow.dataset as ds
except ImportError:
    raise ImportError(
        'pyarrow is not installed. Please install it using `pip install "minari[arrow]"`'
    )

from minari.data_collector.episode_buffer import EpisodeBuffer
from minari.dataset.minari_storage import MinariStorage


class ArrowStorage(MinariStorage):
    FORMAT = "arrow"

    def __init__(
        self,
        data_path: pathlib.Path,
        observation_space: gym.Space,
        action_space: gym.Space,
    ):
        super().__init__(data_path, observation_space, action_space)

    @classmethod
    def _create(
        cls,
        data_path: pathlib.Path,
        observation_space: gym.Space,
        action_space: gym.Space,
    ) -> MinariStorage:
        return cls(data_path, observation_space, action_space)

    def update_episode_metadata(
        self, metadatas: Iterable[Dict], episode_indices: Optional[Iterable] = None
    ):
        if episode_indices is None:
            episode_indices = range(self.total_episodes)

        sentinel = object()
        for new_metadata, episode_id in zip_longest(
            metadatas, episode_indices, fillvalue=sentinel
        ):
            if sentinel in (new_metadata, episode_id):
                raise ValueError("Metadatas and episode_indices have different lengths")

            assert isinstance(new_metadata, dict)
            metadata_path = self.data_path.joinpath(str(episode_id), "metadata.json")

            metadata = {}
            if metadata_path.exists():
                with open(metadata_path) as file:
                    metadata = json.load(file)
            metadata.update(new_metadata)
            with open(metadata_path, "w") as file:
                json.dump(metadata, file, cls=NumpyEncoder)

    def get_episode_metadata(self, episode_indices: Iterable[int]) -> Iterable[Dict]:
        for episode_id in episode_indices:
            metadata_path = self.data_path.joinpath(str(episode_id), "metadata.json")
            with open(metadata_path) as file:
                yield json.load(file)

    def get_episodes(self, episode_indices: Iterable[int]) -> Iterable[dict]:
        dataset = pa.dataset.dataset(
            [
                pa.dataset.dataset(
                    f"{self.data_path}/{ep_id}",
                    format=self.FORMAT,
                    ignore_prefixes=["_", ".", "metadata.json"],
                )
                for ep_id in episode_indices
            ]
        )

        def _to_dict(id, episode):
            return {
                "id": id,
                "observations": _decode_space(
                    self.observation_space, episode["observations"]
                ),
                "actions": _decode_space(self.action_space, episode["actions"][:-1]),
                "rewards": np.asarray(episode["rewards"])[:-1],
                "terminations": np.asarray(episode["terminations"])[:-1],
                "truncations": np.asarray(episode["truncations"])[:-1],
                "infos": _decode_info(episode["infos"])
                if "infos" in episode.column_names
                else {},
            }

        return map(_to_dict, episode_indices, dataset.to_batches())

    def update_episodes(self, episodes: Iterable[EpisodeBuffer]):
        total_steps = self.total_steps
        total_episodes = self.total_episodes
        for episode_data in episodes:
            episode_id = (
                episode_data.id if episode_data.id is not None else total_episodes
            )
            total_episodes = max(total_episodes, episode_id + 1)
            observations = _encode_space(
                self.observation_space, episode_data.observations
            )
            rewards = np.asarray(episode_data.rewards).reshape(-1)
            terminations = np.asarray(episode_data.terminations).reshape(-1)
            truncations = np.asarray(episode_data.truncations).reshape(-1)
            pad = len(observations) - len(rewards)
            actions = _encode_space(self._action_space, episode_data.actions, pad=pad)

            episode_batch = {
                "episode_id": np.full(len(observations), episode_id, dtype=np.int32),
                "observations": observations,
                "actions": actions,
                "rewards": np.pad(rewards, ((0, pad))),
                "terminations": np.pad(terminations, ((0, pad))),
                "truncations": np.pad(truncations, ((0, pad))),
            }
            if episode_data.infos:
                episode_batch["infos"] = _encode_info(episode_data.infos)
            episode_batch = pa.RecordBatch.from_pydict(episode_batch)

            total_steps += len(rewards)
            ds.write_dataset(
                episode_batch,
                self.data_path,
                format=self.FORMAT,
                partitioning=["episode_id"],
                existing_data_behavior="overwrite_or_ignore",
            )

            episode_metadata: dict = {"id": episode_id, "total_steps": len(rewards)}
            if episode_data.seed is not None:
                episode_metadata["seed"] = episode_data.seed
            if episode_data.options is not None:
                episode_metadata["options"] = episode_data.options
            self.update_episode_metadata([episode_metadata], [episode_id])

        self.update_metadata(
            {"total_steps": total_steps, "total_episodes": total_episodes}
        )


def _encode_space(space: gym.Space, values: Any, pad: int = 0):
    if isinstance(space, gym.spaces.Dict):
        assert isinstance(values, dict), values
        arrays, names = [], []
        for key, value in values.items():
            names.append(key)
            arrays.append(_encode_space(space[key], value, pad=pad))
        return pa.StructArray.from_arrays(arrays, names=names)
    if isinstance(space, gym.spaces.Tuple):
        assert isinstance(values, tuple), values
        arrays, names = [], []
        for i, value in enumerate(values):
            names.append(str(i))
            arrays.append(_encode_space(space[i], value, pad=pad))
        return pa.StructArray.from_arrays(arrays, names=names)
    elif isinstance(space, gym.spaces.Box):
        values = np.asarray(values)
        assert values.shape[1:] == space.shape
        values = values.reshape(values.shape[0], -1)
        values = np.pad(values, ((0, pad), (0, 0)))
        dtype = pa.list_(pa.from_numpy_dtype(space.dtype), list_size=values.shape[1])
        return pa.FixedSizeListArray.from_arrays(values.reshape(-1), type=dtype)
    elif isinstance(space, gym.spaces.Discrete):
        values = np.asarray(values).reshape(-1, 1)
        values = np.pad(values, ((0, pad), (0, 0)))
        return pa.array(values.squeeze(-1), type=pa.int32())
    else:
        if not isinstance(values, list):
            values = list(values)
        return pa.array(values + [None] * pad)


def _decode_space(space, values: pa.Array):
    if isinstance(space, gym.spaces.Dict):
        return {
            name: _decode_space(subspace, values.field(name))
            for name, subspace in space.spaces.items()
        }
    elif isinstance(space, gym.spaces.Tuple):
        return tuple(
            [
                _decode_space(subspace, values.field(str(i)))
                for i, subspace in enumerate(space.spaces)
            ]
        )
    elif isinstance(space, gym.spaces.Box):
        data = np.stack(values.to_numpy(zero_copy_only=False))
        return data.reshape(-1, *space.shape)
    elif isinstance(space, gym.spaces.Discrete):
        return values.to_numpy()
    else:
        return values.to_pylist()


def _encode_info(info: dict):
    arrays, fields = [], []

    for key, values in info.items():
        if isinstance(values, dict):
            array = _encode_info(values)
            arrays.append(array)
            fields.append(pa.field(key, array.type))

        elif isinstance(values, tuple):
            array = _encode_info({str(i): v for i, v in enumerate(values)})
            arrays.append(array)
            fields.append(pa.field(key, array.type))

        elif isinstance(values, np.ndarray) or (
            isinstance(values, Sequence) and isinstance(values[0], np.ndarray)
        ):
            if isinstance(values, Sequence):
                values = np.stack(values)

            data_shape = values.shape[1:]
            values = values.reshape(len(values), -1)
            dtype = pa.from_numpy_dtype(values.dtype)
            struct = pa.list_(dtype, list_size=values.shape[1])
            arrays.append(
                pa.FixedSizeListArray.from_arrays(values.reshape(-1), type=struct)
            )
            fields.append(pa.field(key, struct, metadata={"shape": bytes(data_shape)}))

        else:
            array = pa.array(list(values))
            arrays.append(array)
            fields.append(pa.field(key, array.type))

    return pa.StructArray.from_arrays(arrays, fields=fields)


def _decode_info(values: pa.Array):
    nested_dict = {}
    for i, field in enumerate(values.type):
        if isinstance(field, pa.StructArray):
            nested_dict[field.name] = _decode_info(values.field(i))
        else:
            value = np.stack(values.field(i).to_numpy(zero_copy_only=False))
            if field.metadata is not None and b"shape" in field.metadata:
                data_shape = tuple(field.metadata[b"shape"])
                value = value.reshape(len(value), *data_shape)
            nested_dict[field.name] = value
    return nested_dict


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super().default(obj)



import os
from pathlib import Path
from typing import Any, Optional

from minari.storage.remotes.cloud_storage import CloudStorage


try:
    import google.cloud.storage as gcp_storage
    from tqdm import tqdm
except ImportError:
    raise ImportError(
        'google-cloud-storage or tqdm are not installed. Please install it using `pip install "minari[gcs]"`'
    )


class GCPStorage(CloudStorage):
    def __init__(self, name: str, key_path: Optional[str] = None) -> None:
        if key_path is None:
            self.storage_client = gcp_storage.Client.create_anonymous_client()
        else:
            self.storage_client = gcp_storage.Client.from_service_account_json(
                json_credentials_path=key_path
            )
        self.bucket = gcp_storage.Bucket(self.storage_client, name)

    def upload_directory(self, path: Path, remote_dir_path: str) -> None:
        # See https://github.com/googleapis/python-storage/issues/27 for discussion on progress bars
        for local_file in path.glob("*"):
            if local_file.is_dir():
                self.upload_directory(
                    local_file, remote_dir_path + "/" + os.path.basename(local_file)
                )
            else:
                remote_path = os.path.join(remote_dir_path, local_file.name)
                self.upload_file(local_file, remote_path)

    def upload_file(self, local_path: Path, remote_path: str) -> None:
        blob = self.bucket.blob(remote_path)
        blob.upload_from_filename(local_path)

    def list_blobs(self, prefix: Optional[str] = None) -> list:
        return self.bucket.list_blobs(prefix=prefix)

    def download_blob(self, blob: Any, file_path: Path) -> None:
        with open(file_path, "wb") as f:
            with tqdm.wrapattr(f, "write", total=blob.size) as file_obj:
                self.storage_client.download_blob_to_file(blob, file_obj)



from __future__ import annotations

import importlib.metadata
import os
import re
from dataclasses import dataclass, field
from typing import Callable, Iterable, Iterator, List

import gymnasium as gym
import numpy as np
import numpy.typing as npt
from gymnasium import error, logger
from gymnasium.envs.registration import EnvSpec
from packaging.requirements import InvalidRequirement, Requirement
from packaging.version import Version

from minari.data_collector.episode_buffer import EpisodeBuffer
from minari.dataset.episode_data import EpisodeData
from minari.dataset.minari_storage import MinariStorage, PathLike


VERSION_RE = r"(?:-v(?P<version>\d+))"
DATASET_NAME_RE = r"(?:(?P<dataset>[-_\w]+?))"
NAMESPACE_RE = r"(?:(?P<namespace>[-_\w][-_\w/]*[-_\w]+)\/)"
DATASET_ID_RE = re.compile(rf"^{NAMESPACE_RE}?{DATASET_NAME_RE}{VERSION_RE}?$")


def parse_dataset_id(dataset_id: str) -> tuple[str | None, str, int]:
    """Parse dataset ID string format - ``(namespace/)dataset_name(-v[version])``.

    Args:
        dataset_id (str): The dataset id to parse
    Returns:
        A tuple of namespace, dataset name and version number
    Raises:
        Error: If the dataset id is not valid dataset regex
    """
    match = DATASET_ID_RE.fullmatch(dataset_id)
    if not match:
        raise error.Error(
            f"Malformed dataset ID: {dataset_id}. (IDs must be of the form (namespace/)(dataset_name)-v(version). The namespace is optional.)"
        )
    namespace, dataset_name, version = match.group("namespace", "dataset", "version")

    version = int(version)

    if namespace == "":
        namespace = None

    return namespace, dataset_name, version


def gen_dataset_id(
    namespace: str | None,
    dataset_name: str,
    version: int | None = None,
) -> str:
    """Generate a dataset ID from dataset attributes. Inverse of parse_dataset_id().

    Args:
        namespace (str | None): name of dataset subdir. Defaults to None.
        dataset_name (str): name of the dataset.
        version (int | None, optional): Dataset version. Defaults to None, in which case
            the version tag will be suppressed.

    Returns:
        str: A dataset id string of the form ``(namespace/)(dataset_name)(-v(version))``.
            The ``namespace`` and ``-v(version)`` are optional.
    """
    namespace_str = f"{namespace}/" if namespace is not None else ""
    version_str = f"-v{version}" if version is not None else ""
    return f"{namespace_str}{dataset_name}{version_str}"


@dataclass
class MinariDatasetSpec:
    env_spec: EnvSpec | None
    total_episodes: int
    total_steps: int
    dataset_id: str
    combined_datasets: List[str]
    observation_space: gym.Space
    action_space: gym.Space
    data_path: str
    minari_version: str

    # post-init attributes
    namespace: str | None = field(init=False)
    dataset_name: str = field(init=False)
    version: int | None = field(init=False)

    def __post_init__(self):
        """Calls after the spec is created to extract the environment name, dataset name and version from the dataset id."""
        (
            self.namespace,
            self.dataset_name,
            self.version,
        ) = parse_dataset_id(self.dataset_id)


class MinariDataset:
    """Main Minari dataset class to sample data and get metadata information from a dataset."""

    def __init__(
        self,
        data: MinariStorage | PathLike,
        episode_indices: npt.NDArray[np.int_] | None = None,
    ):
        """Initialize properties of the Minari Dataset.

        Args:
            data (Union[MinariStorage, PathLike]): source of data.
            episode_indices (Optional[np.ndarray]): slice of episode indices this dataset is pointing to.
        """
        if isinstance(data, MinariStorage):
            self._data = data
        elif isinstance(data, (str, os.PathLike)):
            self._data = MinariStorage.read(data)
        else:
            raise ValueError(f"Unrecognized type {type(data)} for data")

        self._total_steps = None
        if episode_indices is None:
            episode_indices = np.arange(self._data.total_episodes)
            self._total_steps = self._data.total_steps
        assert episode_indices is not None
        self._episode_indices: npt.NDArray[np.int_] = episode_indices

        metadata = self._data.metadata

        env_spec = metadata.get("env_spec")
        if env_spec is not None:
            assert isinstance(env_spec, str)
            env_spec = (  # for gymnasium 1.0.0 compatibility
                env_spec.replace('"order_enforce": true,', "")
                .replace('"apply_api_compatibility": false,', "")
                .replace('"autoreset": false, ', "")
            )
            env_spec = EnvSpec.from_json(env_spec)
        self._env_spec = env_spec

        eval_env_spec = metadata.get("eval_env_spec")
        if eval_env_spec is not None:
            assert isinstance(eval_env_spec, str)
            eval_env_spec = (  # for gymnasium 1.0.0 compatibility
                eval_env_spec.replace('"order_enforce": true,', "")
                .replace('"apply_api_compatibility": false,', "")
                .replace('"autoreset": false, ', "")
            )
            eval_env_spec = EnvSpec.from_json(eval_env_spec)
        self._eval_env_spec = eval_env_spec

        dataset_id = metadata["dataset_id"]
        assert isinstance(dataset_id, str)
        self._dataset_id = dataset_id

        minari_version = metadata["minari_version"]
        assert isinstance(minari_version, str)
        from minari import __version__, supported_dataset_versions

        if minari_version not in supported_dataset_versions:
            raise ValueError(
                f"The installed Minari version {__version__} does not support the dataset generated by Minari {minari_version}."
                f"Supported versions: {supported_dataset_versions}"
            )
        self._minari_version = minari_version

        self._combined_datasets = metadata.get("combined_datasets", [])

        self._observation_space = metadata["observation_space"]
        self._action_space = metadata["action_space"]
        assert isinstance(self._observation_space, gym.spaces.Space)
        assert isinstance(self._action_space, gym.spaces.Space)

        self._generator = np.random.default_rng()

    def recover_environment(self, eval_env: bool = False, **kwargs) -> gym.Env:
        """Recover the Gymnasium environment used to create the dataset.

        Args:
            eval_env (bool): if True, the returned Gymnasium environment will be that intended to be used for evaluation. If no eval_env was specified when creating the dataset, the returned environment will be the same as the one used for creating the dataset. Default False.
            **kwargs: any other parameter that you want to pass to the `gym.make` function.

        Returns:
            environment: Gymnasium environment
        """
        requirements = self._data.metadata.get("requirements", [])
        for req_str in requirements:
            try:
                req = Requirement(req_str)
            except InvalidRequirement:
                logger.warn(f"Ignoring malformed requirement `{req_str}`")
                continue

            try:
                installed_version = Version(importlib.metadata.version(req.name))
            except importlib.metadata.PackageNotFoundError:
                logger.warn(
                    f'Package {req.name} is not installed. Install it with `pip install "{req_str}"`'
                )
            else:
                if not req.specifier.contains(installed_version):
                    logger.warn(
                        f"Installed {req.name} version {installed_version} does not meet the requirement {req.specifier}.\n"
                        f'We recommend to install the required version with `pip install "{req_str}"`'
                    )

        if eval_env:
            if self._eval_env_spec is not None:
                return gym.make(self._eval_env_spec, **kwargs)
            logger.info(
                f"`eval_env` has been set to True but the dataset {self._dataset_id} doesn't provide an evaluation environment. Instead, the environment used for collecting the data will be returned: {self._env_spec}"
            )

        if self.env_spec is None:
            raise ValueError("Environment cannot be recovered when env_spec is None")

        return gym.make(self.env_spec, **kwargs)

    def set_seed(self, seed: int):
        """Set seed for random episode sampling generator."""
        self._generator = np.random.default_rng(seed)

    def filter_episodes(
        self, condition: Callable[[EpisodeData], bool]
    ) -> MinariDataset:
        """Filter the dataset episodes with a condition.

        The condition must be a callable which takes an `EpisodeData` instance and returns a bool.
        The callable must return a `bool` True if the condition is met and False otherwise.
        i.e filtering for episodes that terminate:

        ```
        dataset.filter(condition=lambda x: x['terminations'][-1] )
        ```

        Args:
            condition (Callable[[EpisodeData], bool]): function that gets in input an EpisodeData object and returns True if certain condition is met.
        """

        def dict_to_episode_data_condition(episode: dict) -> bool:
            return condition(EpisodeData(**episode))

        mask = self.storage.apply(
            dict_to_episode_data_condition, episode_indices=self.episode_indices
        )
        assert self.episode_indices is not None
        filtered_indices = self.episode_indices[list(mask)]
        return MinariDataset(self.storage, episode_indices=filtered_indices)

    def sample_episodes(self, n_episodes: int) -> Iterable[EpisodeData]:
        """Sample n number of episodes from the dataset.

        Args:
            n_episodes (Optional[int], optional): number of episodes to sample.
        """
        indices = self._generator.choice(
            self.episode_indices, size=n_episodes, replace=False
        )
        episodes = self.storage.get_episodes(indices)
        return list(map(lambda data: EpisodeData(**data), episodes))

    def iterate_episodes(
        self, episode_indices: Iterable[int] | None = None
    ) -> Iterator[EpisodeData]:
        """Iterate over episodes from the dataset.

        Args:
            episode_indices (Optional[Iterable[int]], optional): episode indices to iterate over.
        """
        if episode_indices is None:
            assert self.episode_indices is not None
            assert self.episode_indices.ndim == 1
            episode_indices = self.episode_indices

        assert episode_indices is not None
        episodes_data = self.storage.get_episodes(episode_indices)
        return map(lambda data: EpisodeData(**data), episodes_data)

    def update_dataset_from_buffer(self, buffer: List[EpisodeBuffer]):
        """Additional data can be added to the Minari Dataset from a list of episode dictionary buffers.

        Args:
            buffer (list[EpisodeBuffer]): list of episode dictionary buffers to add to dataset
        """
        first_id = self.storage.total_episodes
        self.storage.update_episodes(buffer)
        self.episode_indices = np.append(
            self.episode_indices, first_id + np.arange(len(buffer))
        )

    def __iter__(self):
        return self.iterate_episodes()

    def __getitem__(self, idx: int) -> EpisodeData:
        episode = self.iterate_episodes([self.episode_indices[idx]])
        return next(episode)

    def __len__(self) -> int:
        return self.total_episodes

    @property
    def total_episodes(self) -> int:
        """Total number of episodes in the Minari dataset."""
        return len(self.episode_indices)

    @property
    def total_steps(self) -> int:
        """Total episodes steps in the Minari dataset."""
        if self._total_steps is None:
            self._total_steps = 0
            metadatas = self.storage.get_episode_metadata(self.episode_indices)
            for m in metadatas:
                self._total_steps += m["total_steps"]
        return int(self._total_steps)

    @property
    def episode_indices(self) -> npt.NDArray[np.int_]:
        """Indices of the available episodes to sample within the Minari dataset."""
        return self._episode_indices

    @episode_indices.setter
    def episode_indices(self, new_value: npt.NDArray[np.int_]):
        self._total_steps = None  # invalidate cache
        self._episode_indices = new_value

    @property
    def observation_space(self):
        """Original observation space of the environment before flatteining (if this is the case)."""
        return self._observation_space

    @property
    def action_space(self):
        """Original action space of the environment before flatteining (if this is the case)."""
        return self._action_space

    @property
    def env_spec(self):
        """Envspec of the environment that has generated the dataset."""
        return self._env_spec

    @property
    def combined_datasets(self) -> List[str]:
        """If this Minari dataset is a combination of other subdatasets, return a list with the subdataset names."""
        if self._combined_datasets is None:
            return []
        else:
            return self._combined_datasets

    @property
    def id(self) -> str:
        """Name of the Minari dataset."""
        return self._dataset_id

    @property
    def minari_version(self) -> str:
        """Version of Minari the dataset is compatible with."""
        return self._minari_version

    @property
    def storage(self) -> MinariStorage:
        """Minari storage managing access to disk."""
        return self._data

    @property
    def spec(self) -> MinariDatasetSpec:
        """Minari dataset specifier."""
        return MinariDatasetSpec(
            env_spec=self.env_spec,
            total_episodes=self._episode_indices.size,
            total_steps=self.total_steps,
            dataset_id=self.id,
            combined_datasets=self.combined_datasets,
            observation_space=self.observation_space,
            action_space=self.action_space,
            data_path=str(self.storage.data_path),
            minari_version=str(self.minari_version),
        )



"""Minari CLI commands."""
import os
from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional
from typing_extensions import Annotated

import typer
from gymnasium.envs.registration import EnvSpec
from rich import print
from rich.markdown import Markdown
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from minari import __version__
from minari.dataset.minari_dataset import gen_dataset_id, parse_dataset_id
from minari.storage import get_dataset_path, hosting, local
from minari.utils import combine_datasets, get_dataset_spec_dict, get_env_spec_dict


app = typer.Typer()
list_app = typer.Typer()
app.add_typer(list_app, name="list", short_help="List Minari datasets.")


def _version_callback(value: bool):
    """Show installed Minari version."""
    if value:
        typer.echo(f"Minari version: {__version__}")
        raise typer.Exit()


def _show_dataset_table(datasets: Dict[str, Dict[str, Any]], table_title: str):
    # Collect compatible versions of each dataset
    dataset_versions = defaultdict(list)

    for dataset_id in datasets.keys():
        namespace, dataset_name, version = parse_dataset_id(dataset_id)
        dataset_id_versionless = gen_dataset_id(namespace, dataset_name)
        dataset_versions[dataset_id_versionless].append(version)

    # "Versions" column is only displayed if there are multiple versions
    display_versions = any([len(x) > 1 for x in dataset_versions.values()])

    # Build the Rich Table
    table = Table(title=table_title)

    table.add_column("Name", justify="left", style="cyan", no_wrap=True)

    if display_versions:
        table.add_column("Versions", justify="right", style="green", no_wrap=True)

    table.add_column("Total Episodes", justify="right", style="green", no_wrap=True)
    table.add_column("Total Steps", justify="right", style="green", no_wrap=True)
    table.add_column("Dataset Size", justify="left", style="green", no_wrap=True)
    table.add_column("Author", justify="left", style="magenta", no_wrap=True)

    previous_namespace = None

    for dataset_prefix, versions in dataset_versions.items():
        dataset_id = f"{dataset_prefix}-v{max(versions)}"
        dst_metadata = datasets[dataset_id]
        author = dst_metadata.get("author", "Unknown")
        if isinstance(author, Iterable):
            author = ", ".join(author)
        dataset_size = dst_metadata.get("dataset_size", "Unknown")
        if dataset_size != "Unknown":
            dataset_size = f"{str(dataset_size)} MB"
        author_email = dst_metadata.get("author_email", "Unknown")
        if isinstance(author_email, Iterable):
            author_email = ", ".join(author_email)

        assert isinstance(dst_metadata["dataset_id"], str)
        assert isinstance(author, str)
        assert isinstance(author_email, str)

        docs_url = dst_metadata.get("docs_url", None)
        compatible_versions = ", ".join(
            [f"v{x}" for x in sorted(versions, reverse=True)]
        )

        if docs_url is not None:
            dataset_id_text = f"[link={docs_url}]{dataset_id}[/link]"
        else:
            dataset_id_text = dataset_id

        namespace, _, _ = parse_dataset_id(dataset_id)

        if namespace != previous_namespace:
            table.add_section()
            previous_namespace = namespace

        # Build the current table row
        rows = []
        rows.append(dataset_id_text)

        if display_versions:
            rows.append(compatible_versions)

        rows.append(str(dst_metadata["total_episodes"]))
        rows.append(str(dst_metadata["total_steps"]))
        rows.append(dataset_size)
        rows.append(author)
        table.add_row(*rows)

    print(table)


@app.callback()
def common(
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            "-v",
            callback=_version_callback,
            help="Show installed Minari version.",
        ),
    ] = None,
):
    """Minari is a tool for collecting and hosting Offline datasets for Reinforcement Learning environments based on the Gymnaisum API."""
    pass


@list_app.command("remote")
def list_remote(
    all: Annotated[
        bool, typer.Option("--all", "-a", help="Show all dataset versions.")
    ] = False
):
    """List Minari datasets hosted in the Farama server."""
    if all:
        datasets = hosting.list_remote_datasets()
    else:
        datasets = hosting.list_remote_datasets(
            latest_version=False, compatible_minari_version=True
        )
    table_title = "Minari datasets in Farama server"
    _show_dataset_table(datasets, table_title)


@list_app.command("local")
def list_local(
    all: Annotated[
        bool, typer.Option("--all", "-a", help="Show all dataset versions.")
    ] = False
):
    """List local Minari datasets."""
    if all:
        datasets = local.list_local_datasets()
    else:
        datasets = local.list_local_datasets(
            latest_version=False, compatible_minari_version=True
        )
    dataset_dir = os.environ.get(
        "MINARI_DATASETS_PATH",
        os.path.join(os.path.expanduser("~"), ".minari/datasets/"),
    )
    table_title = f"Local Minari datasets('{dataset_dir}')"
    _show_dataset_table(datasets, table_title)


@app.command()
def show(dataset: Annotated[str, typer.Argument()]):
    """Describe a local or remote dataset, and its environment."""
    local_datasets = local.list_local_datasets()

    # Try to find a local dataset first, then fall back to remote datasets
    if dataset in local_datasets:
        dst_metadata = local_datasets[dataset]
    else:
        remote_datasets = hosting.list_remote_datasets()

        if dataset in remote_datasets:
            dst_metadata = remote_datasets[dataset]
        else:
            local_dataset_path = get_dataset_path("")
            print(
                Text(
                    f"""The dataset `{dataset}` can't be found locally"""
                    f"""(at `{local_dataset_path}`) or remotely.""",
                    style="red",
                )
            )
            raise typer.Abort()

    dataset_id = dst_metadata["dataset_id"]
    description = dst_metadata.get("description")
    docs_url = dst_metadata.get("docs_url")

    if docs_url is not None:
        dataset_id_text = f"[{dataset_id}]({docs_url})"
    else:
        dataset_id_text = dataset_id

    dataset_spec_table = Table(show_header=False, highlight=True)
    dataset_spec_table.add_column(style="bold")
    dataset_spec_table.add_column(style="not bold")

    for key, value in get_dataset_spec_dict(dst_metadata).items():
        md = Markdown(
            str(value), inline_code_lexer="python", inline_code_theme="monokai"
        )
        dataset_spec_table.add_row(key, md)

    print(Markdown(f"""# {dataset_id_text}"""))

    if description is not None:
        print(Markdown(f"""\n## Description\n {description} """))

    print(Markdown("## Dataset Specs"))
    print(dataset_spec_table)

    for env_type in ["env_spec", "eval_env_spec"]:
        env_spec_json = dst_metadata.get(env_type)

        if env_spec_json is not None:
            assert isinstance(env_spec_json, str)
            env_spec_json = (  # for gymnasium 1.0.0 compatibility
                env_spec_json.replace('"order_enforce": true,', "")
                .replace('"apply_api_compatibility": false,', "")
                .replace('"autoreset": false, ', "")
            )
            env_spec = EnvSpec.from_json(env_spec_json)
            env_spec_table = Table(show_header=False, highlight=True)
            env_spec_table.add_column(style="bold")
            env_spec_table.add_column(style="not bold")

            for key, value in get_env_spec_dict(env_spec).items():
                md = Markdown(
                    value, inline_code_lexer="python", inline_code_theme="monokai"
                )
                env_spec_table.add_row(key, md)

            if env_type == "env_spec":
                print(Markdown("## Environment Specs"))
            else:
                print(Markdown("## Evaluation Environment Specs"))

            print(env_spec_table)


@app.command()
def delete(datasets: Annotated[List[str], typer.Argument()]):
    """Delete datasets from local database."""
    # check that the given local datasets exist
    local_dsts = local.list_local_datasets()
    non_matching_local = [dst for dst in datasets if dst not in local_dsts]
    if len(non_matching_local) > 0:
        local_dataset_path = get_dataset_path("")
        tree = Tree(
            f"The following datasets can't be found locally at `{local_dataset_path}`",
            style="red",
        )
        for dst in non_matching_local:
            tree.add(dst, style="magenta")
        print(tree)
        raise typer.Abort()

    # prompt to delete datasets
    datasets_to_delete = {
        local_name: local_dsts[local_name]
        for local_name in local_dsts.keys()
        if local_name in datasets
    }
    _show_dataset_table(datasets_to_delete, "Delete local Minari datasets")
    typer.confirm("Are you sure you want to delete these local datasets?", abort=True)

    for dst in datasets:
        local.delete_dataset(dst)


@app.command()
def download(
    datasets: Annotated[List[str], typer.Argument()],
    force: Annotated[
        bool, typer.Option("--force", "-f", help="Perform a force download.")
    ] = False,
):
    """Download Minari datasets from Farama server."""
    # check if datasets exist in remote server
    remote_dsts = hosting.list_remote_datasets()

    non_matching_remote = [dst for dst in datasets if dst not in remote_dsts]
    if len(non_matching_remote) > 0:
        tree = Tree(
            "The following datasets can't be found in the remote Farama server",
            style="red",
        )
        for dst in non_matching_remote:
            tree.add(dst, style="magenta")
        print(tree)
        raise typer.Abort()

    # check existing datasets locally and ask again if you want them to be overridden
    local_dsts = local.list_local_datasets()
    datasets_to_override = {
        local_name: local_dsts[local_name]
        for local_name in local_dsts.keys()
        if local_name in datasets
    }

    if len(datasets_to_override) > 0 and not force:
        _show_dataset_table(datasets_to_override, "Download remote Minari datasets")
        typer.confirm(
            "Are you sure you want to download and override these local datasets?",
            abort=True,
        )

    # download datastets
    for dst in datasets:
        hosting.download_dataset(dst, force_download=force)


@app.command()
def upload(
    datasets: Annotated[List[str], typer.Argument()],
    key_path: Annotated[str, typer.Option()],
):
    """Upload Minari datasets to the remote Farama server."""
    local_dsts = local.list_local_datasets()
    remote_dsts = hosting.list_remote_datasets()

    # check that the given local datasets exist
    non_matching_local = [dst for dst in datasets if dst not in local_dsts]
    if len(non_matching_local) > 0:
        local_dataset_path = get_dataset_path("")
        tree = Tree(
            f"The following datasets can't be found locally at `{local_dataset_path}`",
            style="red",
        )
        for dst in non_matching_local:
            tree.add(dst, style="magenta")
        print(tree)
        raise typer.Abort()

    # check that non of the datasets exist in the Farama server
    matching_remote = {
        remote_name: remote_dsts[remote_name]
        for remote_name in remote_dsts.keys()
        if remote_name in datasets
    }
    if len(matching_remote) > 0:
        print(
            "[red]The following datasets are already present in the Farama server, please contact the Farama team at contact@farama.org to upload your datasets.[/red]",
        )
        _show_dataset_table(matching_remote, "Matching datasets in Farama server")
        raise typer.Abort()

    # Upload datasets
    for dst in datasets:
        hosting.upload_dataset(dst, key_path)


@app.command()
def combine(
    datasets: Annotated[List[str], typer.Argument()],
    dataset_id: Annotated[str, typer.Option()],
):
    """Combine multiple datasets into a single Minari dataset."""
    local_dsts = local.list_local_datasets()
    # check dataset name doesn't exist locally
    if dataset_id in local_dsts:
        print(
            f"[red]Dataset name {dataset_id} already exist in the local Minari datasets.[/red]",
        )
        raise typer.Abort()

    # check list of local datasets to combine exist
    non_matching_local = [dst for dst in datasets if dst not in local_dsts]
    if len(non_matching_local) > 0:
        local_dataset_path = get_dataset_path("")
        tree = Tree(
            f"The following datasets can't be found locally at `{local_dataset_path}`",
            style="red",
        )
        for dst in non_matching_local:
            tree.add(dst, style="magenta")
        print(tree)
        raise typer.Abort()
    if len(datasets) > 1:
        minari_datasets = list(map(lambda x: local.load_dataset(x), datasets))
        combine_datasets(minari_datasets, dataset_id)
        print(
            f"The datasets [green]{datasets}[/green] were successfully combined into [blue]{dataset_id}[/blue]!"
        )
    else:
        print(
            f"[red]The list of local datasets to combine {datasets} must be of size two or greater.[/red]",
        )
        raise typer.Abort()


if __name__ == "__main__":
    app()



import os
from pathlib import Path


def get_dataset_path(dataset_id: str) -> Path:
    """Get the path to a dataset main directory."""
    datasets_path = os.environ.get("MINARI_DATASETS_PATH")
    if datasets_path is not None:
        file_path = os.path.join(datasets_path, dataset_id)
    else:
        datasets_path = os.path.join(os.path.expanduser("~"), ".minari", "datasets")
        file_path = os.path.join(datasets_path, dataset_id)

    os.makedirs(datasets_path, exist_ok=True)
    return Path(file_path)



import pathlib

import pytest
from mktestdocs import check_md_file


def test_readme():
    fpath = pathlib.Path("README.md")
    check_md_file(fpath=fpath, lang="bash")
    check_md_file(fpath=fpath)


@pytest.mark.parametrize("fpath", [*pathlib.Path("docs").glob("**/*.md")], ids=str)
def test_markdown(fpath):
    check_md_file(fpath=fpath, lang="bash")
    check_md_file(fpath=fpath)



import os
from typing import Callable, Dict, Type

from minari.storage.remotes.cloud_storage import CloudStorage


DEFAULT_REMOTE = "gcp://minari-remote"


def get_gcps() -> Type[CloudStorage]:
    from .gcp import GCPStorage

    return GCPStorage


_registry: Dict[str, Callable[[], Type[CloudStorage]]] = {"gcp": get_gcps}


def get_cloud_storage(key_path=None) -> CloudStorage:
    remote_spec = os.getenv("MINARI_REMOTE", DEFAULT_REMOTE)
    cloud_type, name = remote_spec.split("://", maxsplit=1)
    return _registry[cloud_type]()(name, key_path)
